import pytest_asyncio

from bson import ObjectId

from motordantic.document import Document
from motordantic.session import SessionSync


class TicketSync(Document):
    name: str
    position: int
    config: dict
    sign: int = 1
    type_: str = 'ga'
    array: list = [1, 2]

    class Config:
        excluded_query_fields = ('sign', 'type')


@pytest_asyncio.fixture(scope='session', autouse=True)
async def drop_ticket_sync_collection(event_loop):
    yield
    await TicketSync.Q.drop_collection(force=True)


def test_sync_insert_one(connection):
    data = {'name': 'sync1', 'position': 2310, 'config': {'as_sync': True}}
    object_id = TicketSync.Qsync.insert_one(**data)
    assert isinstance(object_id, ObjectId)


def test_sync_find_one(connection):
    ticket = TicketSync.Qsync.find_one(name='sync1')
    assert ticket.name == 'sync1'
    assert ticket.position == 2310


def test_sync_insert_many(connection):
    data = [
        {'name': 'sync2', 'position': 2, 'config': {'as_sync': True}},
        {'name': 'sync3', 'position': 3, 'config': {'as_sync': True}},
        {'name': 'sync_fourth', 'position': 4, 'config': {'as_sync': False}},
    ]
    inserted = TicketSync.Qsync.insert_many(data)
    assert inserted == 3


def test_sync_aggregation(connection):
    summ = TicketSync.Qsync.aggregate_sum('position')
    assert summ == 2319

    max_ = TicketSync.Qsync.aggregate_max('position')
    assert max_ == 2310

    min_ = TicketSync.Qsync.aggregate_min('position')
    assert min_ == 2

    avg = TicketSync.Qsync.aggregate_avg('position')
    assert avg == 579.75


def test_find_sync(connection):
    data = TicketSync.Qsync.find().list
    assert len(data) == 4
    assert data[-1].name == 'sync_fourth'


def test_sync_distinct(connection):
    data = TicketSync.Qsync.distinct('position', name='sync2')
    assert data == [2]


def test_session_sync(connection):
    with SessionSync(TicketSync.manager) as session:
        ticket = TicketSync.Qsync.find_one(session=session)
        assert ticket is not None



import os

from motordantic.connection import (
    connect,
    MotordanticConnection,
)

from motor.motor_asyncio import AsyncIOMotorClient


class TestWriteConnectionParams:
    def setup(self):
        connection = connect("mongodb://127.0.0.1:27017", "test")
        self.connection: MotordanticConnection = connection

    def test_envirnament_values(self):
        database_name = os.getenv('MOTORDANTIC_DATABASE')
        assert database_name == 'test'

        address = os.getenv('MOTORDANTIC_ADDRESS')
        assert address == "mongodb://127.0.0.1:27017"

    def test_connection_params(self):
        database_name = self.connection.database_name
        conn_string = self.connection.address
        assert database_name == "test"
        assert conn_string == "mongodb://127.0.0.1:27017"

    def test_connection(self):
        motor_client = self.connection._get_motor_client()
        assert isinstance(motor_client, AsyncIOMotorClient)

    def test_conection_database(self):
        motor_client = self.connection._get_motor_client()
        assert motor_client.get_database('test') == AsyncIOMotorClient(
            "mongodb://127.0.0.1:27017"
        ).get_database("test")



import pytest
import pytest_asyncio

from motordantic.document import Document
from motordantic.query.query import Q


class TicketForQuery(Document):
    name: str
    position: int


@pytest_asyncio.fixture(scope='session', autouse=True)
async def drop_ticket_collection(event_loop):
    yield
    await TicketForQuery.Q.drop_collection(force=True)


def test_query_organization(connection):
    query = Q(name=123) | Q(name__ne=124) & Q(position=1) | Q(position=2)
    data = query.to_query(TicketForQuery.manager._builder)
    value = {
        '$or': [
            {'name': '123'},
            {'$and': [{'name': {'$ne': '124'}}, {'position': 1}]},
            {'position': 2},
        ]
    }
    assert data == value


@pytest.mark.asyncio
async def test_query_result(connection):
    query = [
        TicketForQuery(name='first', position=1),
        TicketForQuery(name='second', position=2),
    ]
    inserted = await TicketForQuery.Q.insert_many(query)
    assert inserted == 2

    query = Q(name='first') | Q(position=1) & Q(name='second')
    data = await TicketForQuery.Q.find_one(query)
    assert data.name == 'first'

    query = Q(position=3) | Q(position=1) & Q(name='second')
    data = await TicketForQuery.Q.find_one(query)
    assert data is None

    query = Q(position=3) | Q(position=2) & Q(name='second')
    data = await TicketForQuery.Q.find_one(query)
    assert data.name == 'second'



from typing import TYPE_CHECKING, Any, Callable, Dict, Optional, Type, cast

from pydantic.fields import Field as PydanticField
from pydantic.fields import FieldInfo as PydanticFieldInfo
from pydantic.fields import ModelField, Representation, Undefined

__all__ = (
    'ExtraDBField',
    'MotordanticFieldInfo',
    'ComputedField',
    'computed_field',
)

if TYPE_CHECKING:
    from typing import Callable, Union

    from pydantic import BaseConfig, BaseModel
    from pydantic.class_validators import Validator
    from pydantic.fields import ModelField


class MotordanticFieldInfo:
    __slots__ = ("pydantic_field_info", "db_field_name", "default")

    def __init__(
        self,
        *,
        pydantic_field_info: PydanticFieldInfo,
        db_field_name: Optional[str],
    ):
        self.pydantic_field_info = pydantic_field_info
        self.db_field_name = db_field_name


def ExtraDBField(
    default: Any = Undefined,
    *,
    db_field_name: str,
    default_factory: Optional[Callable[[], Any]] = None,
    title: Optional[str] = None,
    description: Optional[str] = None,
    const: Optional[bool] = None,
    gt: Optional[float] = None,
    ge: Optional[float] = None,
    lt: Optional[float] = None,
    le: Optional[float] = None,
    multiple_of: Optional[float] = None,
    min_items: Optional[int] = None,
    max_items: Optional[int] = None,
    min_length: Optional[int] = None,
    max_length: Optional[int] = None,
    regex: Optional[str] = None,
    **extra: Any,
) -> Any:
    pydantic_field = PydanticField(
        default,
        default_factory=default_factory,
        # alias=db_field_name,
        title=cast(str, title),
        description=cast(str, description),
        const=cast(bool, const),
        gt=cast(float, gt),
        ge=cast(float, ge),
        lt=cast(float, lt),
        le=cast(float, le),
        multiple_of=cast(float, multiple_of),
        min_items=cast(int, min_items),
        max_items=cast(int, max_items),
        min_length=cast(int, min_length),
        max_length=cast(int, max_length),
        regex=cast(str, regex),
        **extra,
    )
    return MotordanticFieldInfo(
        pydantic_field_info=pydantic_field,
        db_field_name=db_field_name,
    )


def computed_field(
    func: Optional[Any] = None,
    *,
    alias: Optional[str] = None,
    title: Optional[str] = None,
    description: Optional[str] = None,
) -> 'Union[Callable[[Any], ComputedField], ComputedField]':
    import inspect

    def wrapper(func_: Any) -> 'ComputedField':
        if not isinstance(func_, property):
            func_ = property(func_)
        type_ = inspect.signature(func_.fget).return_annotation  # type: ignore
        if type_ is inspect._empty:  # type: ignore[attr-defined]
            type_ = Any

        return ComputedField(
            name=func_.fget.__name__,  # type: ignore
            type_=type_,
            fget=func_.fget,  # type: ignore
            alias=alias,
            title=title,
            description=description,
        )

    if func is None:
        return wrapper

    return wrapper(func)


class ComputedField(Representation):
    __slots__ = (
        'name',
        'type_',
        'outer_type_',
        'fget',
        'fset',
        'class_validators',
        'alias',
        'title',
        'description',
        'config',
        'model_field',
        'required',
    )

    def __init__(
        self,
        *,
        name: str,
        type_: type,
        fget: Callable[[Optional['BaseModel']], Any],
        fset: Any = None,
        alias: Optional[str] = None,
        title: Optional[str] = None,
        description: Optional[str] = None,
        config: Optional[Type['BaseConfig']] = None,
        model_field: Optional['ModelField'] = None,
    ) -> None:
        self.name: str = name
        self.type_: type = type_
        self.outer_type_: type = type_
        self.fget: Callable[[Optional['BaseModel']], Any] = fget
        self.fset: Any = fset
        self.class_validators: Optional[Dict[str, 'Validator']] = None

        self.alias: str = alias or name
        self.title: Optional[str] = title
        self.description: Optional[str] = description

        self.config: Optional[Type['BaseConfig']] = config
        self.model_field: Optional[ModelField] = model_field
        self.required: bool = False

    @property
    def field_info(self) -> PydanticFieldInfo:
        return PydanticFieldInfo(
            alias=self.alias,
            title=self.title,
            description=self.description or self.fget.__doc__,
            read_only=True,
        )

    def __get__(self, instance: Optional['BaseModel'], owner: Any) -> Any:
        if instance is None:
            return self
        return self.fget(instance)

    def __set__(self, obj, value):
        raise AttributeError("can't set attribute")

    def setter(self, fset):
        return type(self)(
            name=self.name,
            type_=self.type_,
            fget=self.fget,
            fset=fset,
            alias=self.alias,
            title=self.title,
            description=self.description,
            config=self.config,
            model_field=self.model_field,
        )

    def set_config_and_prepare_field(self, config: Type['BaseConfig']) -> None:
        self.config = config
        self.model_field = ModelField(
            name=self.name,
            type_=self.type_,
            class_validators=self.class_validators,
            model_config=self.config,
            default=None,
            required=self.required,
            alias=self.field_info.alias,  # type: ignore
            field_info=self.field_info,
        )

    @classmethod
    def __get_validators__(cls):
        yield



import inspect
from typing import (
    Optional,
    List,
    Any,
    Dict,
    Union,
    List,
    Tuple,
    Generator,
    TYPE_CHECKING,
)
from re import compile, IGNORECASE

from bson import ObjectId, Regex
from pymongo import UpdateOne
from pydantic.fields import ModelField
from pydantic.typing import get_origin

from ..property import cached_classproperty
from ..validation import validate_field_value, validate_object_id
from ..exceptions import MotordanticValidationError
from ..types import Relation, RelationInfo, RelationTypes

__all__ = (
    'ExtraQueryMapper',
    'group_by_aggregate_generation',
    'generate_name_field',
    'chunk_by_length',
    'bulk_query_generator',
    'take_relation',
)

if TYPE_CHECKING:
    from ..document import Document


class ExtraQueryMapper(object):
    """extra mapper for __ queries like find(_id__in=[], name__regex='123')"""

    __slots__ = (
        'document',
        'field_name',
    )

    def __init__(self, document: 'Document', field_name: str):
        self.field_name = field_name
        self.document = document

    def query(self, extra_methods: List, values: Any) -> Dict:
        if self.field_name == '_id':
            values = (
                [validate_object_id(self.document, v) for v in values]
                if isinstance(values, list)
                else validate_object_id(self.document, values)
            )
        if extra_methods:
            query: Dict = {self.field_name: {}}
            for extra_method in extra_methods:
                if extra_method == 'in':
                    extra_method = 'in_'
                elif extra_method == 'inc':
                    return self.inc(values)
                elif extra_method == 'unset':
                    return self.unset(values)
                query[self.field_name].update(getattr(self, extra_method)(values))
            return query
        return {}

    def in_(self, list_values: List) -> dict:
        if not isinstance(list_values, list):
            raise TypeError("values must be a list type")
        try:
            return {
                "$in": [
                    validate_field_value(self.document, self.field_name, v)
                    for v in list_values
                ]
            }
        except MotordanticValidationError:
            return {"$in": list_values}

    def regex(self, regex_value: str) -> dict:
        return {"$regex": Regex.from_native(compile(regex_value))}

    def iregex(self, regex_value: str) -> dict:
        return {"$regex": Regex.from_native(compile(regex_value, IGNORECASE))}

    def regex_ne(self, regex_value: str) -> dict:
        return {"$not": Regex.from_native(compile(regex_value))}

    def ne(self, value: Any) -> dict:
        return {"$ne": validate_field_value(self.document, self.field_name, value)}

    def startswith(self, value: str) -> dict:
        return {"$regex": Regex.from_native(compile(f"^{value}"))}

    def istartswith(self, value: str) -> dict:
        return {"$regex": Regex.from_native(compile(f"^{value}", IGNORECASE))}

    def not_startswith(self, value: str) -> dict:
        return {"$not": Regex.from_native(compile(f"^{value}"))}

    def not_istartswith(self, value: str) -> dict:
        return {"$not": Regex.from_native(compile(f"^{value}", IGNORECASE))}

    def endswith(self, value: str) -> dict:
        return {"$regex": Regex.from_native(compile(f"{value}$"))}

    def iendswith(self, value: str) -> dict:
        return {"$regex": Regex.from_native(compile(f"{value}$", IGNORECASE))}

    def not_endswith(self, value: str) -> dict:
        return {"$not": Regex.from_native(compile(f"{value}$"))}

    def nin(self, list_values: List) -> dict:
        if not isinstance(list_values, list):
            raise TypeError("values must be a list type")
        try:
            return {
                "$nin": [
                    validate_field_value(self.document, self.field_name, v)
                    for v in list_values
                ]
            }
        except MotordanticValidationError:
            return {"$nin": list_values}

    def exists(self, boolean_value: bool) -> dict:
        if not isinstance(boolean_value, bool):
            raise TypeError("boolean_value must be a bool type")
        return {"$exists": boolean_value}

    def type(self, bson_type) -> dict:
        return {"$type": bson_type}

    def search(self, search_text: str) -> dict:
        return {'$search': search_text}

    def all(self, query: Any) -> dict:
        return {'$all': query}

    def unset(self, value: Any) -> dict:
        return {"$unset": {self.field_name: value}}

    def gte(self, value: Any) -> dict:
        return {"$gte": validate_field_value(self.document, self.field_name, value)}

    def lte(self, value: Any) -> dict:
        return {"$lte": validate_field_value(self.document, self.field_name, value)}

    def gt(self, value: Any) -> dict:
        return {"$gt": validate_field_value(self.document, self.field_name, value)}

    def lt(self, value: Any) -> dict:
        return {"$lt": validate_field_value(self.document, self.field_name, value)}

    def inc(self, value: int) -> dict:
        if isinstance(value, int):
            return {'$inc': {self.field_name: value}}
        raise ValueError('value must be integer')

    def range(self, range_values: Union[List, Tuple]) -> dict:
        if len(range_values) != 2:
            raise ValueError("range must have 2 params")
        from_ = range_values[0]
        to_ = range_values[1]
        return {
            "$gte": validate_field_value(self.document, self.field_name, from_),
            "$lte": validate_field_value(self.document, self.field_name, to_),
        }

    @cached_classproperty
    def methods(cls) -> list:
        methods = []
        for f in cls.__dict__:
            if f == 'in_':
                methods.append('in')
            elif not f.startswith('__') and f != 'query':
                methods.append(f)
        return methods


def group_by_aggregate_generation(
    group_by: Union[str, list, tuple]
) -> Union[str, dict]:
    """group by parametr generation helper"""

    if isinstance(group_by, (list, tuple)):
        return {
            g if '.' not in g else g.split('.')[-1]: f'${g}' if '$' not in g else g
            for g in group_by
        }
    if '.' in group_by:
        name = group_by.split('.')[-1]
        return {name: f'${group_by}'}
    return f'${group_by}' if not '$' in group_by else group_by


def generate_name_field(name: Union[dict, str, None] = None) -> Optional[str]:
    if isinstance(name, dict):
        return '|'.join(str(v) for v in name.values())
    return name


def chunk_by_length(items: Union[List, Tuple], step: int) -> Generator:
    """Yield successive n-sized chunks from l."""
    for i in range(0, len(items), step):
        yield items[i : i + step]


def bulk_query_generator(
    requests: List,
    updated_fields: Optional[List] = None,
    query_fields: Optional[List] = None,
    upsert=False,
) -> List:
    """ "helper for generate bulk query"""

    data = []
    if updated_fields:
        for obj in requests:
            query = {'_id': ObjectId(obj._id)}
            update = {}
            for field in updated_fields:
                value = getattr(obj, field)
                update.update({field: value})
            data.append(UpdateOne(query, {'$set': update}, upsert=upsert))
    elif query_fields:
        for obj in requests:
            query = {}
            update = {}
            for field, value in obj.data.items():
                if field not in query_fields:
                    update.update({field: value})
                else:
                    query.update({field: value})
            data.append(UpdateOne(query, {'$set': update}, upsert=upsert))
    return data


def take_relation(field: ModelField) -> Optional[RelationInfo]:
    """take relations from model

    Args:
        field (ModelField): Document field

    Returns:
        Optional[RelationInfo]: relation info
    """
    if field.type_ == Relation:
        if field.allow_none is True:
            return RelationInfo(
                field=field.name,
                document_class=field.sub_fields[0].type_,  # type: ignore
                relation_type=RelationTypes.OPTIONAL_SINGLE,
            )
        return RelationInfo(
            field=field.name,
            document_class=field.sub_fields[0].type_,  # type: ignore
            relation_type=RelationTypes.SINGLE,
        )
    if (
        inspect.isclass(get_origin(field.outer_type_))
        and issubclass(get_origin(field.outer_type_), list)  # type: ignore
        and len(field.sub_fields) == 1  # type: ignore
    ):
        internal_field = field.sub_fields[0]  # type: ignore
        if internal_field.type_ == Relation:
            if internal_field.allow_none is True:
                return None
            return RelationInfo(
                field=field.name,
                document_class=internal_field.sub_fields[0].type_,  # type: ignore
                relation_type=RelationTypes.ARRAY,
            )
    return None



import uuid

import pytest_asyncio
import pytest

from bson import ObjectId

from motordantic.document import Document
from motordantic.types import UUIDField
from motordantic.exceptions import MotordanticValidationError


class Article(Document):
    article_id: UUIDField
    title: str
    tags: list


@pytest_asyncio.fixture(scope='session', autouse=True)
async def innert_users(event_loop):
    yield
    await Article.Q.drop_collection(force=True)


@pytest.mark.asyncio
async def test_insert_one_with_uuid_field(connection):
    with pytest.raises(MotordanticValidationError):
        article = Article(article_id='2310', title='invalid', tags=['invalid'])

    uid = uuid.uuid1()
    article = Article(article_id=uid, title='invalid', tags=['invalid'])
    await article.save()
    assert isinstance(article._id, ObjectId)
    assert isinstance(article.article_id, uuid.UUID)
    assert article.article_id == uid

    db_article = await Article.Q.find_one(article_id=uid)
    assert db_article._id == article._id

    db_article_hex = await Article.Q.find_one(article_id=uid.hex)
    assert db_article_hex._id == db_article._id



from .extra import (
    ExtraQueryMapper,
    generate_name_field,
    group_by_aggregate_generation,
    bulk_query_generator,
)
from .query import Q, QCombination, generate_basic_query, QueryBuilder
from .result import FindResult, SimpleAggregateResult, AggregateResult



from typing import TYPE_CHECKING, Any, Optional, Dict, List, Union

from ..exceptions import MotordanticValidationError
from ..query import generate_basic_query, Q, QCombination, AggregateResult

__all__ = ('Aggregate',)

if TYPE_CHECKING:
    from .document import Document


class Aggregate(object):
    def __init__(self, document_class: 'Document'):
        self.document_class = document_class
        self.pipeline: List[Dict] = []

    def match(
        self, logical_query: Optional[Union[Q, QCombination]] = None, **query
    ) -> 'Aggregate':
        if logical_query is not None:
            query_params = self.document_class.Q._check_query_args(logical_query)  # type: ignore
        else:
            query_params = generate_basic_query(
                self.document_class.manager, query, False
            )
        self.pipeline.append({'$match': {**query_params}})
        return self

    def raw_match(self, native_match_query: dict) -> 'Aggregate':
        self.pipeline.append({'$match': native_match_query})
        return self

    def add_fields(self, **fields) -> 'Aggregate':
        self.pipeline.append({'$addFields': {**fields}})
        return self

    def group(self, _id: Any = None, **group_kwargs) -> 'Aggregate':
        self.pipeline.append({'$group': {'_id': _id, **group_kwargs}})
        return self

    def project(self, project_args: dict) -> 'Aggregate':
        self.pipeline.append({'$project': {**project_args}})
        return self

    def skip(self, skip: int) -> 'Aggregate':
        self.pipeline.append({'$skip': skip})
        return self

    def limit(self, limit: int) -> 'Aggregate':
        self.pipeline.append({'$limit': limit})
        return self

    def sort(self, **sort_kwargs) -> 'Aggregate':
        self.pipeline.append({'$sort': {**sort_kwargs}})
        return self

    def lookup(
        self,
        from_: 'Document',
        local_field: str,
        foreign_field: str,
        as_: Optional[str] = None,
    ) -> 'Aggregate':
        if local_field not in self.document_class.__fields__ and local_field != '_id':
            raise MotordanticValidationError(
                f'field - {local_field} not a field from model: {from_.__name__}'
            )
        if foreign_field not in from_.__fields__ and foreign_field != '_id':
            raise MotordanticValidationError(
                f'field - {foreign_field} not a field from model: {self.document_class.__name__}'
            )
        lookup = {
            'from': from_.get_collection_name(),
            'localField': local_field,
            'foreignField': foreign_field,
        }
        if as_:
            lookup['as'] = as_
        else:
            lookup['as'] = (
                f'{from_.__name__.lower()}es'
                if from_.__name__.endswith('s')
                else f'{from_.__name__.lower()}s'
            )
        self.pipeline.append({'$lookup': lookup})
        return self

    def unwind(self, unwind: Union[str, dict]) -> 'Aggregate':
        if isinstance(unwind, str) and not unwind.startswith('$'):
            raise MotordanticValidationError('unwind must be startswith $')
        self.pipeline.append({'$unwind': unwind})
        return self

    def facet(self, facet: Dict[str, 'Aggregate']) -> 'Aggregate':
        assert any([not isinstance(v, Aggregate) for v in facet.values()])
        self.pipeline.append(
            {'$facet': {name: aggregate.pipeline for name, aggregate in facet.items()}}
        )
        return self

    def out(self, name: str) -> 'Aggregate':
        self.pipeline.append({'$out': name})
        return self

    def replace_with(self, value: Union[str, dict]) -> 'Aggregate':
        if isinstance(value, str) and not value.startswith('$'):
            raise MotordanticValidationError('string value must be startwith $')
        self.pipeline.append({'$replaceWith': value})
        return self

    def redact(self, redact: Dict[str, Any]) -> 'Aggregate':
        self.pipeline.append({'$redact': redact})
        return self

    def replace_root(self, new_root: Union[str, dict]) -> 'Aggregate':
        if isinstance(new_root, str) and not new_root.startswith('$'):
            raise MotordanticValidationError('string new_root must be startwith $')
        self.pipeline.append({'$replaceRoot': {"newRoot": new_root}})
        return self

    def sample(self, **sample) -> 'Aggregate':
        self.pipeline.append({'$sample': {**sample}})
        return self

    def fill(
        self,
        sort_by: dict,
        output: dict,
        partition_by: Optional[dict] = None,
        partition_by_fields: Optional[list] = None,
    ) -> 'Aggregate':
        fill = {
            'sortBy': sort_by,
            'output': output,
        }
        if partition_by:
            fill['partitionBy'] = partition_by
        if partition_by_fields:
            fill['partitionByFields'] = partition_by_fields  # type: ignore
        self.pipeline.append({'$fill': fill})
        return self

    def geo_near(
        self,
        distance_field: str,
        near: dict,
        distance_multiplier: Optional[Union[int, float]] = None,
        include_locs: Optional[str] = None,
        key: Optional[str] = None,
        max_distance: Optional[float] = None,
        min_distance: Optional[float] = None,
        query: Optional[dict] = None,
        spherical: Optional[bool] = None,
    ) -> 'Aggregate':
        geo_near: Dict[str, Any] = {'distanceField': distance_field, 'near': near}
        if distance_multiplier is not None:
            geo_near['distanceMultiplier'] = distance_multiplier
        if include_locs:
            geo_near['includeLocs'] = include_locs
        if key:
            geo_near['key'] = key
        if max_distance:
            geo_near['maxDistance'] = max_distance
        if min_distance:
            geo_near['maxDistance'] = min_distance
        if query:
            geo_near['query'] = query
        if spherical is not None:
            geo_near['spherical'] = spherical
        self.pipeline.append({'$geoNear': geo_near})
        return self

    def let(self, let: dict) -> 'Aggregate':
        self.pipeline.append({'$let': {**let}})
        return self

    async def result(self) -> AggregateResult:
        result = await self.document_class.Q.raw_aggregate(self.pipeline)
        return AggregateResult(native_result=result, document_class=self.document_class)

    def result_sync(self) -> AggregateResult:
        result = self.document_class.Qsync.raw_aggregate(self.pipeline)
        return AggregateResult(native_result=result, document_class=self.document_class)



import pytest
import pytest_asyncio

from motordantic.document import Document
from motordantic.fields import computed_field
from motordantic.exceptions import MotordanticValidationError
from pydantic import BaseModel


class Config(BaseModel):
    path: str = '/home/'
    env: str = 'test'


class Application(Document):
    name: str
    config: Config
    lang: str

    @computed_field
    def lang_upper(self) -> str:
        return self.lang.upper()


@pytest_asyncio.fixture(scope='session', autouse=True)
async def application_data(connection):
    application = await Application(name='test', config=Config(), lang='python').save()
    yield
    await Application.Q.drop_collection(force=True)


def test_schema(connection):
    application = Application(name='test', config=Config(), lang='python')
    assert application.serialize(['lang_upper']) == {'lang_upper': 'PYTHON'}
    assert application.serialize_json(['lang_upper']) == '{"lang_upper": "PYTHON"}'
    assert application.data == {
        'name': 'test',
        'config': {'path': '/home/', 'env': 'test'},
        'lang': 'python',
        'lang_upper': 'PYTHON',
    }
    assert (
        application.json()
        == '{"name": "test", "config": {"path": "/home/", "env": "test"}, "lang": "python", "lang_upper": "PYTHON"}'
    )
    assert application.schema() == {
        'title': 'Application',
        'type': 'object',
        'properties': {
            'name': {'title': 'Name', 'type': 'string'},
            'config': {'$ref': '#/definitions/Config'},
            'lang': {'title': 'Lang', 'type': 'string'},
            'lang_upper': {'title': 'Lang Upper', 'read_only': True, 'type': 'string'},
        },
        'required': ['name', 'config', 'lang'],
        'definitions': {
            'Config': {
                'title': 'Config',
                'type': 'object',
                'properties': {
                    'path': {'title': 'Path', 'default': '/home/', 'type': 'string'},
                    'env': {'title': 'Env', 'default': 'test', 'type': 'string'},
                },
            }
        },
    }


@pytest.mark.asyncio
async def test_application_data(connection):
    application = await Application.Q.find_one(name='test')
    data = application.data
    assert isinstance(data['config'], dict)
    assert data['config']['env'] == 'test'

    data = await Application.Q.find_one(config__env='test')
    assert data.name == 'test'
    data = await Application.Q.find_one(config__env='invalid')
    assert data is None


@pytest.mark.asyncio
async def test_raise_with_field_mongo_model(connection):
    class Default(Document):
        name: str
        app: Application

    with pytest.raises(MotordanticValidationError):
        await Default(name='default', app=await Application.Q.find_one()).save()



import pytest_asyncio
import pytest

from bson import ObjectId

from motordantic.document import Document
from motordantic.fields import ExtraDBField
from motordantic.exceptions import NotDeclaredField


class FieldDoc(Document):
    doc_id: str = ExtraDBField(db_field_name='documentId')
    name: str = ExtraDBField(db_field_name='documentName')


class AggDoc(Document):
    doc_id: str = ExtraDBField(db_field_name='documentId')
    cost: float = ExtraDBField(db_field_name='price')


@pytest_asyncio.fixture(scope='session', autouse=True)
async def drop_field_doc(event_loop, connection):
    yield
    await FieldDoc.Q.drop_collection(force=True)
    await AggDoc.Q.drop_collection(force=True)


@pytest.mark.asyncio
async def test_save(connection):
    doc = FieldDoc(doc_id='123', name='first document')
    await doc.save()
    assert doc._id is not None

    native_result = await FieldDoc.manager.collection.find_one({'_id': doc._id})
    assert native_result.get('documentId') == doc.doc_id
    assert native_result.get('documentName') == doc.name


@pytest.mark.asyncio
async def test_find_one(connection):
    doc = await FieldDoc.Q.find_one(doc_id='123')
    assert doc is not None
    assert doc.doc_id == '123'
    assert doc.name == 'first document'

    with pytest.raises(NotDeclaredField):
        doc = await FieldDoc.Q.find_one(documentId='123')
        assert doc == 1


@pytest.mark.asyncio
async def test_insert_one(connection):
    data = {'doc_id': '2222', 'name': 'second'}
    object_id = await FieldDoc.Q.insert_one(**data)
    assert isinstance(object_id, ObjectId)


@pytest.mark.asyncio
async def test_update_in_save(connection):
    doc = await FieldDoc.Q.find_one(doc_id__regex='2222')
    assert doc is not None
    doc.name = 'second document'
    await doc.save()
    native_result = await FieldDoc.manager.collection.find_one({'_id': doc._id})
    assert native_result.get('documentName') == 'second document'


@pytest.mark.asyncio
async def test_simple_aggregate(connection):

    data = [{'doc_id': '1', 'cost': 5000}, {'doc_id': '2', 'cost': 6000}]
    c = await AggDoc.Q.insert_many(data)
    assert c == 2
    sum_ = await AggDoc.Q.aggregate_sum('cost')
    assert sum_ == 11000

    max_ = await AggDoc.Q.aggregate_max('cost')
    assert max_ == 6000
    min_ = await AggDoc.Q.aggregate_min('cost')
    assert min_ == 5000
    avg_ = await AggDoc.Q.aggregate_avg('cost')
    assert avg_ == 5500.0



import pytest
import pytest_asyncio

from pymongo.collection import WriteConcern

from motordantic.document import Document


class Box(Document):
    name: str
    position: int


@pytest_asyncio.fixture(scope='session', autouse=True)
async def drop_ticket_collection(event_loop):
    yield
    await Box.Q.drop_collection(force=True)


@pytest.mark.asyncio
async def test_insert_many_with_wrice_concern(connection):
    data = [
        {
            'name': 'first',
            'position': 1,
        },
        {
            'name': 'second',
            'position': 2,
        },
    ]
    write_concern = WriteConcern(w=0)
    inserted_count = await Box.Q.insert_many(
        data, ordered=False, write_concern=write_concern
    )
    assert inserted_count == 2



from typing import TYPE_CHECKING, Dict, Any
from dataclasses import dataclass

from ..exceptions import MotordanticValidationError

__all__ = ('Sum', 'Avg', 'Min', 'Count', 'Max')

if TYPE_CHECKING:
    from ..document import Document


class BaseAggregationOperation(object):
    """Abstract class for Aggregation"""

    __slots__ = ('field', '_operation')

    _operation: str

    def __init__(self, field: str):
        self.field = field

    @property
    def operation(self) -> str:
        if not self._operation:
            raise NotImplementedError('implement _operation')
        return self._operation

    def _validate_field(self, document_class: 'Document'):
        # if self.field in document_class.__motordantic_extra_fields__.values():
        #     return None
        if self.field not in document_class.__fields__ and self.field != '_id':
            raise MotordanticValidationError(
                f'invalid field "{self.field}" for this model, field must be one of {list(document_class.__fields__.keys())}'
            )

    def _aggregate_query(self, document_class: 'Document') -> dict:
        self._validate_field(document_class)
        query_field = document_class.__mapping_query_fields__[self.field]
        query = {
            f'{query_field}__{self.operation}': {f'${self.operation}': f'${self.field}'}
        }
        return query


class Sum(BaseAggregationOperation):
    """
    Simple sum aggregation

    generated query: {'field__sum': {'$sum': 'field'}}

    return: {'field__sum': <value>}
    """

    _operation: str = 'sum'


class Max(BaseAggregationOperation):
    """
    Simple max aggregation

    generated query: {'field__max': {'$max': 'field'}}

    return: {'field__max': <value>}
    """

    _operation: str = 'max'


class Min(BaseAggregationOperation):
    """
    Simple min aggregation

    generated query: {'field__min': {'$min': 'field'}}

    return: {'field__min': <value>}
    """

    _operation: str = 'min'


class Avg(BaseAggregationOperation):
    """
    Simple avg aggregation

    generated query: {'field__avg': {'$avg': 'field'}}

    return: {'field__avg': <value>}
    """

    _operation: str = 'avg'


class Count(BaseAggregationOperation):
    """
    Simple Count aggregation

    generated query:
        - if field = _id
            {
                '_id': None
                'count': {'$sum': 1},
            }
        - else
            {S
                '_id': 'field'
                'count': {'$sum': 1},
            }
    return: {'_id': <value>, 'count': value}
    """

    _operation: str = 'count'

    def _aggregate_query(self, document_class: 'Document') -> dict:
        self._validate_field(document_class)
        query_field = document_class.__mapping_query_fields__[self.field]
        query = {
            "_id": f'${query_field}' if self.field != '_id' else None,
            f'count': {'$sum': 1},
        }
        return query


@dataclass
class Bucket(object):
    group_by: str
    boundaries: list
    default: str
    output: Dict[str, Any]

    def to_mongo(self) -> Dict[str, Any]:
        bucket = {
            'groupBy': self.group_by,
            'boundaries': self.boundaries,
            'default': self.default,
            'output': self.output,
        }
        return bucket



__all__ = (
    'cached_classproperty',
    'classproperty',
)


class cached_classproperty(classmethod):
    def __init__(self, fget):
        self.obj = {}
        self.fget = fget

    def __get__(self, owner, cls):
        if cls in self.obj:
            return self.obj[cls]
        self.obj[cls] = self.fget(cls)
        return self.obj[cls]


class classproperty(classmethod):
    def __init__(self, method=None):
        self.fget = method

    def __get__(self, instance, cls=None):
        return self.fget(cls)  # type: ignore

    def getter(self, method):
        self.fget = method
        return self



from typing import Optional, TYPE_CHECKING, List
import asyncio

from motor.core import AgnosticClientSession, AgnosticCollection, AgnosticDatabase
from pymongo.errors import AutoReconnect, ServerSelectionTimeoutError, NetworkTimeout


from .query import QueryBuilder, ExtraQueryMapper
from .query.builder import Builder
from .exceptions import NotDeclaredField
from .relation import RelationManager

from .sync import SyncQueryBuilder
from .aggregate.aggregate import Aggregate

if TYPE_CHECKING:
    from .document import Document
    from .connection import MotordanticConnection


class ODMManager(object):
    __database__: Optional[AgnosticDatabase] = None
    __collection__: Optional[AgnosticCollection] = None
    __connection__: Optional['MotordanticConnection'] = None
    __relation_manager__: Optional['RelationManager'] = None

    def __init__(self, document: 'Document'):
        self._builder: Builder = Builder(self)
        self.querybuilder: QueryBuilder = QueryBuilder(self._builder)
        self.sync_querybuilder: SyncQueryBuilder = SyncQueryBuilder(self._builder)
        self.__document__ = document
        if self.__document__.has_relations:
            self.__relation_manager__ = RelationManager(self.__document__)
        else:
            self.__relation_manager__ = None

    @property
    def database(self) -> AgnosticDatabase:
        """Returns the database that is currently associated with this document."""
        if not hasattr(self, "__database__") or self.__database__ is None:
            raise AttributeError("Accessing database without using it first.")
        return self.__database__

    @property
    def connection(self) -> 'MotordanticConnection':
        assert self.__connection__ is not None
        return self.__connection__

    @property
    def motor_client(self):
        return self.connection._get_motor_client()

    @property
    def _io_loop(self) -> asyncio.AbstractEventLoop:
        try:
            _io_loop = (
                self.motor_client.io_loop
                if self.motor_client.io_loop is not None
                else asyncio.get_running_loop()
            )
        except RuntimeError:
            _io_loop = asyncio.new_event_loop()
        return _io_loop

    @property
    def collection(self) -> AgnosticCollection:
        """Returns the collection for this :class:`Document`."""
        if (
            self.__collection__ is None
            or self.__collection__.database is not self.database
        ):
            self.__collection__ = self.database.get_collection(self.document.get_collection_name())  # type: ignore
        return self.__collection__

    async def _start_session(self) -> AgnosticClientSession:
        return await self.motor_client.start_session()

    @property
    def relation_manager(self) -> Optional[RelationManager]:
        return self.__relation_manager__

    @classmethod
    def use(cls, connection: 'MotordanticConnection') -> None:
        assert connection is not None
        cls.__connection__ = connection
        cls.__database__ = connection._get_motor_client().get_database(
            connection.database_name
        )

    @property
    def document(self) -> 'Document':
        assert self.__document__ is not None
        return self.__document__

    def aggregate(self) -> Aggregate:
        aggregate = Aggregate(self.__document__)
        return aggregate

    def _validate_field(self, field: str) -> bool:
        # if field in self.document.__mapping_from_fields__:
        #     return True
        if field not in self.document.__fields__ and field != '_id':
            raise NotDeclaredField(field, list(self.document.__fields__))
        elif field in self.document.__database_exclude_fields__:
            return False
        return True

    def _parse_extra_params(self, extra_params: List) -> tuple:
        field_param, extra = [], []
        methods = ExtraQueryMapper.methods
        for param in extra_params:
            if param in methods:
                extra.append(param)
            else:
                field_param.append(param)
        return field_param, extra

    async def ensure_indexes(self):
        """method for create/update/delete indexes if indexes declared in Config property"""

        indexes = self.document.__indexes__
        if indexes:
            db_indexes = await self.querybuilder.list_indexes()
            indexes_to_create = [
                i for i in indexes if i.document['name'] not in db_indexes  # type: ignore
            ]
            indexes_to_delete = [
                i
                for i in db_indexes
                if i not in [i.document['name'] for i in indexes] and i != '_id_'  # type: ignore
            ]
            result = []
            if indexes_to_create:
                try:
                    result = await self.querybuilder.create_indexes(indexes_to_create)
                except (AutoReconnect, ServerSelectionTimeoutError, NetworkTimeout):
                    pass
            if indexes_to_delete:
                for index_name in indexes_to_delete:
                    await self.querybuilder.drop_index(index_name)
                db_indexes = await self.querybuilder.list_indexes()
            indexes = set(list(db_indexes.keys()) + result)



import pytest_asyncio
import pytest

from motordantic.document import Document


class InnerTicket(Document):
    name: str
    position: int
    config: dict
    params: dict
    sign: int = 1
    type_: str = 'ga'
    array: list = []

    class Config:
        excluded_query_fields = ('sign', 'type')


@pytest_asyncio.fixture(scope='session', autouse=True)
async def innert_tickets(event_loop):

    await InnerTicket.Q.insert_one(
        name='first',
        position=1,
        config={'url': 'localhost', 'username': 'admin'},
        params={},
        array=['1', 'type'],
    )
    await InnerTicket.Q.insert_one(
        name='second',
        position=2,
        config={'url': 'google.com', 'username': 'staff'},
        params={'1': 2},
        array=['second', 'type2'],
    )
    await InnerTicket.Q.insert_one(
        name='third',
        position=3,
        config={'url': 'yahoo.com', 'username': 'trololo'},
        params={'1': 1},
        array=['third', 'type3'],
    )
    await InnerTicket.Q.insert_one(
        name='fourth',
        position=4,
        config={'url': 'yahoo.com', 'username': 'trololo'},
        params={'2': 2},
        array=['fourth', 'type4'],
    )
    yield
    await InnerTicket.Q.drop_collection(force=True)


@pytest.mark.asyncio
async def test_update_many(connection):
    # .create_documents()
    updated = await InnerTicket.Q.update_many(
        position__range=[3, 4], name__ne='hhh', config__url__set='test.io'
    )
    assert updated == 2
    last = await InnerTicket.Q.find_one(position=4)
    assert last.config['url'] == 'test.io'


@pytest.mark.asyncio
async def test_inner_find_one_in_array(connection):
    data = await InnerTicket.Q.find_one(array__1__regex='2')
    assert data.name == 'second'

    data = await InnerTicket.Q.find_one(array__1__regex='2', array__0__regex='1')
    assert data is None

    data = await InnerTicket.Q.find_one(array__0='1')
    assert data.name == 'first'


@pytest.mark.asyncio
async def test_inner_find_one(connection):
    data = await InnerTicket.Q.find_one(config__url__startswith='google', params__1=2)
    assert data.name == 'second'

    data = await InnerTicket.Q.find_one(
        config__url__startswith='yahoo', params__1='qwwe'
    )
    assert data is None


@pytest.mark.asyncio
async def test_inner_update_one(connection):
    updated = await InnerTicket.Q.update_one(
        config__url__startswith='goo', config__url__set='test.io'
    )
    assert updated == 1
    data = await InnerTicket.Q.find_one(config__url__startswith='test')
    assert data.name == 'second'



import pytest
import pytest_asyncio
from pymongo import IndexModel

from motordantic.document import Document
from motordantic.exceptions import MotordanticIndexError


class IndexTicket(Document):
    name: str
    position: int
    config: dict

    class Config:
        indexes = [
            IndexModel([('position', 1)]),
            IndexModel([('name', 1)]),
        ]


@pytest_asyncio.fixture(scope='session', autouse=True)
async def drop_ticket_collection(event_loop):
    await IndexTicket.ensure_indexes()
    yield
    await IndexTicket.Q.drop_collection(force=True)


@pytest.mark.asyncio
async def test_check_indexes(connection):
    result = await IndexTicket.Q.list_indexes()
    assert result == {
        '_id_': {'key': {'_id': 1}},
        'position_1': {'key': {'position': 1}},
        'name_1': {'key': {'name': 1}},
    }


@pytest.mark.asyncio
async def test_check_indexes_if_remove(connection):
    class IndexTicket(Document):
        name: str
        position: int
        config: dict

        class Config:
            indexes = [
                IndexModel([('position', 1)]),
            ]

    await IndexTicket.ensure_indexes()
    result = await IndexTicket.Q.list_indexes()
    assert result == {
        '_id_': {'key': {'_id': 1}},
        'position_1': {'key': {'position': 1}},
    }


@pytest.mark.asyncio
async def test_drop_index(connection):
    with pytest.raises(MotordanticIndexError):
        result = await IndexTicket.Q.drop_index('position1111')

    result = await IndexTicket.Q.drop_index('position_1')
    assert result == 'position_1 dropped.'



from typing import TYPE_CHECKING

from contextlib import ContextDecorator

if TYPE_CHECKING:
    from .manager import ODMManager

__all__ = ('Session', 'SessionSync')


class Session(ContextDecorator):
    __slots__ = ('_session', 'odm_manager')

    def __init__(self, odm_manager: 'ODMManager'):
        self.odm_manager = odm_manager

    async def __aenter__(self):
        self._session = await self.odm_manager._start_session()
        return self._session

    async def __aexit__(self, *args, **kwargs):
        await self._session.end_session()  # type: ignore

    async def start_stransaction(
        self,
        read_concern=None,
        write_concern=None,
        read_preference=None,
        max_commit_time_ms=None,
    ):
        # TODO tranaction login with rs
        return self._session.start_transaction(
            read_preference=read_preference,
            write_concern=write_concern,
            read_concern=read_concern,
            max_commit_time_ms=max_commit_time_ms,
        )


class SessionSync(ContextDecorator):
    __slots__ = ('_session', 'odm_manager')

    def __init__(self, odm_manager: 'ODMManager'):
        self.odm_manager = odm_manager

    def __enter__(self):
        self._session = self.odm_manager._io_loop.run_until_complete(
            self.odm_manager._start_session()
        )
        return self._session

    def __exit__(self, *args, **kwargs):
        return self.odm_manager._io_loop.run_until_complete(self._session.end_session())  # type: ignore



from typing import (
    TYPE_CHECKING,
    AbstractSet,
    Any,
    Dict,
    List,
    Mapping,
    Set,
    Type,
    Union,
)

__all__ = (
    'DocumentType',
    'DictStrList',
    'DictStrAny',
    'DictAny',
    'SetStr',
    'ListStr',
    'IntStr',
    'AbstractSetIntStr',
    'DictIntStrAny',
    'ExcludeInclude',
    'MappingIntStrAny',
)

if TYPE_CHECKING:
    from .document import Document

IntStr = Union[int, str]
DictIntStrAny = Dict[IntStr, Any]
MappingIntStrAny = Mapping[IntStr, Any]
DocumentType = Type['Document']
DictStrList = Dict[str, List]
DictStrAny = Dict[str, Any]
DictAny = Dict[Any, Any]
SetStr = Set[str]
ListStr = List[str]
AbstractSetIntStr = AbstractSet[IntStr]
ExcludeInclude = Union[AbstractSetIntStr, MappingIntStrAny, Any]



import json
from typing import (
    Dict,
    Any,
    Union,
    Optional,
    List,
    Tuple,
    TYPE_CHECKING,
    ClassVar,
)

from bson import ObjectId, DBRef, decode as bson_decode
from bson.raw_bson import RawBSONDocument
from motor.core import AgnosticClientSession
from pydantic.main import ModelMetaclass as PydanticModelMetaclass, default_ref_template
from pydantic import (
    BaseModel as BasePydanticModel,
    root_validator,
    BaseConfig,
    ValidationError,
)
from pydantic.main import ValueItems, _missing
from pydantic.typing import resolve_annotations
from pydantic.fields import Undefined
from pymongo import IndexModel


from .relation import RelationManager
from .types import ObjectIdStr, RelationInfo, Relation
from .exceptions import (
    MotordanticValidationError,
    MotordanticConnectionError,
)
from .property import classproperty
from .query.extra import take_relation
from .fields import ComputedField, MotordanticFieldInfo
from .schema import model_schema
from .manager import ODMManager


__all__ = ('Document', 'DocumentMetaclass')

if TYPE_CHECKING:
    from asyncio import AbstractEventLoop
    from pydantic.main import MappingIntStrAny, TupleGenerator
    from .typing import DictStrAny, AbstractSetIntStr, SetStr
    from .query import QueryBuilder
    from .sync.query import SyncQueryBuilder


_is_document_class_defined = False


class DocumentMetaclass(PydanticModelMetaclass):
    def __new__(mcs, name, bases, namespace, **kwargs):  # type: ignore
        annotations = resolve_annotations(
            namespace.get("__annotations__", {}), namespace.get("__module__")
        )
        mapping_query_fields = {'_id': '_id'}
        for field_name in annotations:
            value = namespace.get(field_name, Undefined)
            if isinstance(value, MotordanticFieldInfo):
                namespace[field_name] = value.pydantic_field_info
                mapping_query_fields[field_name] = value.db_field_name or field_name
            else:
                mapping_query_fields[field_name] = field_name
        namespace['__mapping_query_fields__'] = mapping_query_fields
        namespace['__mapping_from_fields__'] = {
            db_field: cls_field for cls_field, db_field in mapping_query_fields.items()
        }
        cls = super().__new__(mcs, name, bases, namespace, **kwargs)
        indexes = set()
        config = BaseConfig
        if _is_document_class_defined and issubclass(cls, Document):
            db_refs = {}
            for k, v in cls.__fields__.items():
                relation_info = take_relation(v)
                if relation_info is not None:
                    db_refs[k] = relation_info
            setattr(cls, '__db_refs__', db_refs)
            if db_refs:
                setattr(cls, 'has_relations', True)
                setattr(cls, '__relation_manager__', RelationManager(cls))  # type: ignore
            computed_fields = {}
            for var_name, value in namespace.items():
                if isinstance(value, ComputedField):
                    computed_fields[var_name] = value
                    value.set_config_and_prepare_field(config)
            if computed_fields:
                setattr(cls, '__motordantic_computed_fields__', computed_fields)
        json_encoders = getattr(cls.Config, 'json_encoders', {})  # type: ignore
        json_encoders.update({ObjectId: lambda f: str(f)})
        setattr(cls.Config, 'json_encoders', json_encoders)  # type: ignore
        exclude_fields = getattr(cls.Config, 'exclude_fields', tuple())  # type: ignore
        collection_name = (
            getattr(cls.Config, 'collection_name', None) or cls.__name__.lower()
        )
        setattr(cls, '__collection_name__', collection_name)
        setattr(cls, '__indexes__', indexes)
        setattr(cls, '__database_exclude_fields__', exclude_fields)
        setattr(cls, '__manager__', ODMManager(cls))  # type: ignore
        return cls


class Document(BasePydanticModel, metaclass=DocumentMetaclass):
    __indexes__: 'SetStr' = set()
    __manager__: ODMManager
    __database_exclude_fields__: Union[Tuple, List] = tuple()
    __db_refs__: ClassVar[Optional[Dict[str, RelationInfo]]] = None
    __relation_manager__: Optional[RelationManager] = None
    __motordantic_computed_fields__: Dict[str, ComputedField] = {}
    __mapping_query_fields__: Dict[str, str] = {}
    __mapping_from_fields__: Dict[str, str] = {}
    __collection_name__: Optional[str] = None
    _id: Optional[ObjectIdStr] = None
    has_relations: ClassVar[bool] = False

    def __init__(self, **data):
        try:
            super().__init__(**data)
        except ValidationError as e:
            raise MotordanticValidationError(e.errors, e)

    def __setattr__(self, key, value):
        if key == '_id':
            self.__dict__[key] = value
            return value
        else:
            return super().__setattr__(key, value)

    def __getattribute__(self, attr):
        attribute = super().__getattribute__(attr)
        if isinstance(attribute, ComputedField):
            return attribute.fget(self)
        return attribute

    @property
    def _io_loop(self) -> 'AbstractEventLoop':
        return self.manager._io_loop

    @classmethod
    async def ensure_indexes(cls):
        """method for create/update/delete indexes if indexes declared in Config property"""

        indexes = getattr(cls.__config__, 'indexes', [])
        if not all([isinstance(index, IndexModel) for index in indexes]):
            raise ValueError('indexes must be list of IndexModel instances')
        if indexes:
            db_indexes = await cls.Q.list_indexes()
            indexes_to_create = [
                i for i in indexes if i.document['name'] not in db_indexes
            ]
            indexes_to_delete = [
                i
                for i in db_indexes
                if i not in [i.document['name'] for i in indexes] and i != '_id_'
            ]
            result = []
            if indexes_to_create:
                try:
                    result = await cls.Q.create_indexes(indexes_to_create)
                except MotordanticConnectionError:
                    pass
            if indexes_to_delete:
                for index_name in indexes_to_delete:
                    await cls.Q.drop_index(index_name)
                db_indexes = await cls.Q.list_indexes()
            indexes = set(list(db_indexes.keys()) + result)
        setattr(cls, '__indexes__', indexes)

    @classmethod
    def _get_properties(cls) -> list:
        return [
            prop
            for prop in dir(cls)
            if prop
            not in (
                "__values__",
                "data",
                "querybuilder",
                "Q",
                "Qsync",
                "pk",
                "_query_data",
                "_mongo_query_data",
                "fields_all",
                "_io_loop",
            )
            and isinstance(getattr(cls, prop), property)
        ]

    @classmethod
    def parse_obj(cls, data: Any) -> Any:
        if cls.__motordantic_computed_fields__:
            data = {
                k: v
                for k, v in data.items()
                if k not in cls.__motordantic_computed_fields__
            }
        obj = super().parse_obj(data)
        if '_id' in data:
            obj._id = data['_id'].__str__()
        return obj

    async def save(
        self,
        updated_fields: Union[Tuple, List] = [],
        session: Optional[AgnosticClientSession] = None,
    ) -> Any:
        if self._id is not None:
            data = {
                '_id': (
                    self._id if isinstance(self._id, ObjectId) else ObjectId(self._id)
                )
            }
            if updated_fields:
                if not all(field in self.__fields__ for field in updated_fields) or any(
                    field in self.__motordantic_computed_fields__
                    for field in updated_fields
                ):
                    raise MotordanticValidationError('invalid field in updated_fields')
            else:
                updated_fields = tuple(self.__fields__.keys())
            for field in updated_fields:
                if field in self.__motordantic_computed_fields__:
                    continue
                else:
                    data[f'{field}__set'] = getattr(self, field)
            await self.Q.update_one(
                session=session,
                **data,
            )
            return self
        data = {
            field: value
            for field, value in self.__dict__.items()
            if field in self.__fields__
        }
        object_id = await self.Q.insert_one(
            session=session,
            **data,
        )
        self._id = object_id
        return self

    def save_sync(
        self,
        updated_fields: Union[Tuple, List] = [],
        session: Optional[AgnosticClientSession] = None,
    ):
        return self._io_loop.run_until_complete(self.save(updated_fields, session))

    async def delete(self) -> None:
        await self.Q.delete_one(_id=self.pk)

    def delete_sync(self) -> None:
        return self._io_loop.run_until_complete(self.delete())

    @classproperty
    def fields_all(cls) -> list:
        """return all fields with properties(not document fields)"""
        fields = list(cls.__fields__.keys())
        return_fields = fields + cls._get_properties()
        return return_fields

    @classproperty
    def manager(cls) -> ODMManager:
        return cls.__manager__

    @classproperty
    def Q(cls) -> 'QueryBuilder':
        return cls.manager.querybuilder

    @classproperty
    def Qsync(cls) -> 'SyncQueryBuilder':
        return cls.manager.sync_querybuilder

    def _iter(
        self,
        to_dict: bool = False,
        by_alias: bool = False,
        include: Optional[Union['AbstractSetIntStr', 'MappingIntStrAny']] = None,
        exclude: Optional[Union['AbstractSetIntStr', 'MappingIntStrAny']] = None,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
    ) -> 'TupleGenerator':

        # Merge field set excludes with explicit exclude parameter with explicit overriding field set options.
        # The extra "is not None" guards are not logically necessary but optimizes performance for the simple case.
        if exclude is not None or self.__exclude_fields__ is not None:
            exclude = ValueItems.merge(self.__exclude_fields__, exclude)

        if include is not None or self.__include_fields__ is not None:
            include = ValueItems.merge(self.__include_fields__, include, intersect=True)

        allowed_keys = self._calculate_keys(
            include=include, exclude=exclude, exclude_unset=exclude_unset  # type: ignore
        )
        if allowed_keys is None and not (
            to_dict or by_alias or exclude_unset or exclude_defaults or exclude_none
        ):
            # huge boost for plain _iter()
            yield from self.__dict__.items()
            return

        value_exclude = ValueItems(self, exclude) if exclude is not None else None
        value_include = ValueItems(self, include) if include is not None else None

        for field_key, v in self.__dict__.items():
            if (allowed_keys is not None and field_key not in allowed_keys) or (
                exclude_none and v is None
            ):
                continue

            if exclude_defaults:
                model_field = self.__fields__.get(field_key)
                if (
                    not getattr(model_field, 'required', True)
                    and getattr(model_field, 'default', _missing) == v
                ):
                    continue

            if by_alias and field_key in self.__fields__:
                dict_key = self.__fields__[field_key].alias
            else:
                dict_key = field_key

            if to_dict or value_include or value_exclude:
                v = self._get_value(
                    v,
                    to_dict=to_dict,
                    by_alias=by_alias,
                    include=value_include and value_include.for_element(field_key),
                    exclude=value_exclude and value_exclude.for_element(field_key),
                    exclude_unset=exclude_unset,
                    exclude_defaults=exclude_defaults,
                    exclude_none=exclude_none,
                )
            yield dict_key, v

        for field_key, v in self.__motordantic_computed_fields__.items():
            yield field_key, v.fget(self)

    def dict(  # type: ignore
        self,
        *,
        include: Optional['AbstractSetIntStr'] = None,
        exclude: Optional['AbstractSetIntStr'] = None,
        by_alias: bool = False,
        skip_defaults: Optional[bool] = None,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        with_props: bool = True,
    ) -> 'DictStrAny':
        """
        Generate a dictionary representation of the model, optionally specifying which fields to include or exclude.

        """
        attribs = super().dict(
            include=include,  # type: ignore
            exclude=exclude,  # type: ignore
            by_alias=by_alias,
            skip_defaults=skip_defaults,  # type: ignore
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
        )
        if self.__motordantic_computed_fields__:
            for field, value in self.__motordantic_computed_fields__.items():
                attribs[field] = value.fget(self)
        if with_props:
            props = self._get_properties()
            # Include and exclude properties
            if include:
                props = [prop for prop in props if prop in include]
            if exclude:
                props = [prop for prop in props if prop not in exclude]

            # Update the attribute dict with the properties
            if props:
                attribs.update({prop: getattr(self, prop) for prop in props})
        else:
            if self.__motordantic_computed_fields__:
                for field in self.__motordantic_computed_fields__:
                    attribs.pop(field)
        if self.has_relations:
            for field in self.__db_refs__:  # type: ignore
                attrib_data = attribs[field]
                if attrib_data and not isinstance(attrib_data, dict):
                    attribs[field] = (
                        attrib_data.to_dict()
                        if not isinstance(attrib_data, list)
                        else [
                            a.to_dict() if not isinstance(a, dict) else a
                            for a in attrib_data
                        ]
                    )
        return attribs

    @classmethod
    def schema(
        cls, by_alias: bool = True, ref_template: str = default_ref_template  # type: ignore
    ) -> 'DictStrAny':
        cached = cls.__schema_cache__.get((by_alias, ref_template))
        if cached is not None:
            return cached
        s = model_schema(cls, by_alias=by_alias, ref_template=ref_template)
        cls.__schema_cache__[(by_alias, ref_template)] = s
        return s

    @classmethod
    def from_bson(cls, bson_raw_data: RawBSONDocument) -> 'Document':
        data = bson_decode(bson_raw_data.raw)
        data = {
            cls.__mapping_from_fields__[field]: value for field, value in data.items()
        }
        obj = cls(**data)
        obj._id = data.get('_id')
        return obj

    @classmethod
    def to_db_ref(cls, object_id: Union[str, ObjectId]) -> DBRef:
        if isinstance(object_id, str):
            object_id = ObjectId(object_id)
        db_ref = DBRef(collection=cls.get_collection_name(), id=object_id)
        return db_ref

    @classmethod
    def to_relation(cls, object_id: Union[str, ObjectId]) -> Relation:
        db_ref = cls.to_db_ref(object_id=object_id)
        return Relation(db_ref, cls)

    @property
    def data(self) -> 'DictStrAny':
        return self.dict(with_props=True)

    @property
    def _query_data(self) -> 'DictStrAny':
        return self.dict(with_props=False)

    @property
    def _mongo_query_data(self) -> 'DictStrAny':
        return self._query_data

    @classmethod
    def get_collection_name(cls) -> str:
        """main method for set collection

        Returns:
            str: collection name
        """
        return cls.__collection_name__ or cls.__name__.lower()

    def serialize(self, fields: Union[Tuple, List]) -> 'DictStrAny':
        data: dict = self.dict(include=set(fields))
        return {f: data[f] for f in fields}

    def serialize_json(self, fields: Union[Tuple, List]) -> str:
        return json.dumps(self.serialize(fields))

    @property
    def pk(self):
        return self._id

    @root_validator
    def validate_all_fields(cls, values):
        for field, value in values.items():
            if isinstance(value, Document):
                raise ValueError(
                    f'{field} - cant be instance of Document without Relation'
                )
        return values


_is_document_class_defined = True



from typing import List, Optional

import pytest_asyncio
import pytest

from motordantic.document import Document
from motordantic.types import Relation


class Author(Document):
    name: str


class Book(Document):
    title: str
    author: Relation[Author]


class Publish(Document):
    name: str
    books: List[Relation[Book]]


class OptionalTestPublisher(Document):
    name: str
    author: Optional[Relation[Author]] = None


@pytest_asyncio.fixture(scope='session', autouse=True)
async def relation_data(event_loop, connection):
    authors = [Author(name='Author1'), Author(name='Author2')]
    await Author.Q.insert_many(authors)

    author1 = await Author.Q.find_one(name='Author1')
    author2 = await Author.Q.find_one(name='Author2')
    book1 = Book(
        title='first book from author1',
        author=author1,
    )
    book2 = await Book(
        title='first book from author2',
        author=author2,
    ).save()
    book1_id = await Book.Q.insert_one(**book1.data)

    book = await Book.Q.find_one(title__regex='2')
    publish1 = Publish(name='publush1', books=[book])
    await publish1.save()
    publish2 = Publish(name='publush2', books=[Book.to_relation(book1_id), book2])
    await publish2.save()
    optional_publisher = await OptionalTestPublisher.Q.insert_one(
        name='test_publisher1'
    )
    optional_publisher = await OptionalTestPublisher.Q.insert_one(
        name='test_publisher2', author=author2
    )
    yield
    await Author.Q.drop_collection(force=True)
    await Book.Q.drop_collection(force=True)
    await Publish.Q.drop_collection(force=True)
    await OptionalTestPublisher.Q.drop_collection(force=True)


@pytest.mark.asyncio
async def test_book_relation(connection):
    book = await Book.Q.find_one(title='first book from author1')
    book_author = await book.author.get()
    native_author = await Author.Q.find_one(name='Author1')
    assert book_author == native_author

    book = await Book.Q.find_one(
        title='first book from author1', with_relations_objects=True
    )
    # assert book_author == 1
    assert book.author == native_author


@pytest.mark.asyncio
async def test_publish_relation(connection):
    publish = await Publish.Q.find_one(name='publush1', with_relations_objects=True)

    assert publish.data['books'][0]['title'] == 'first book from author2'
    book = await Book.Q.find_one(
        title='first book from author2', with_relations_objects=True
    )
    assert publish.books == [book]
    publish2 = await Publish.Q.find_one(name__regex='2', with_relations_objects=True)
    assert len(publish2.books) == 2
    assert isinstance(publish2.books[0], Book)


@pytest.mark.asyncio
async def test_find_relation_models_without_relations_objects(connection):
    book = await Book.Q.find_one(title__regex='1')
    assert book.author.db_ref.collection == 'author'
    assert isinstance(book.data['author'], dict)


@pytest.mark.asyncio
async def test_create_object(connection):
    author = Author(name='J. R. R. Tolkien')
    await author.save()
    book = Book(title='lord of the rings', author=author.data)
    await book.save()

    assert isinstance(book.author, Relation)
    book = await Book.Q.find_one(_id=book._id)
    assert book.title == 'lord of the rings'
    relation_author = await book.author.get()
    assert relation_author.name == 'J. R. R. Tolkien'


@pytest.mark.asyncio
async def test_optional_publisher(connection):
    optional_publisher = await OptionalTestPublisher.Q.find_one(
        name='test_publisher1', with_relations_objects=True
    )
    assert optional_publisher.author is None

    optional_publisher = await OptionalTestPublisher.Q.find_one(
        name='test_publisher2', with_relations_objects=True
    )
    assert optional_publisher.author is not None
    assert optional_publisher.author.name == 'Author2'



import asyncio
import functools
from typing import Callable

__all__ = ('force_sync',)


def force_sync(fn: Callable, loop=None):
    '''
    turn an async function to sync function
    '''

    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        res = fn(*args, **kwargs)
        if asyncio.iscoroutine(res):
            try:
                loop = asyncio.get_running_loop()
            except RuntimeError:
                loop = asyncio.new_event_loop()
            return loop.run_until_complete(res)
        return res

    return wrapper


def add_sync_version(func):
    # assert asyncio.iscoroutine(func)

    def wrapper(*args, **kwds):
        return asyncio.new_event_loop().run_until_complete(func, *args, **kwds)

    func.sync = wrapper
    return func



import pytest_asyncio
import pytest

from motordantic.document import Document
from motordantic.types import ObjectIdStr
from motordantic.query.query import Q
from motordantic.aggregate.expressions import (
    Sum,
    Max,
    Min,
    Avg,
    Count,
)
from motordantic.exceptions import MotordanticValidationError


class Product(Document):
    title: str
    cost: float
    quantity: int
    product_type: str
    config: dict


class ProductImage(Document):
    url: str
    product_id: ObjectIdStr


@pytest_asyncio.fixture(scope='session', autouse=True)
async def innert_producs(event_loop):
    from random import randint

    product_types = {1: 'phone', 2: 'book', 3: 'food'}
    data = [
        Product(
            title=str(i),
            cost=float(i),
            quantity=i,
            product_type=product_types[2] if i != 4 else product_types[1],
            config={'type_id': i},
        )
        for i in range(1, 5)
    ]
    await Product.Q.insert_many(data)
    products = [p._id for p in await Product.Q.find()]
    for product_id in products:
        for i in range(1, randint(2, 6)):
            _ = await ProductImage(
                url=f'https://image.com/image-{i}',
                product_id=product_id,
            ).save()
    yield
    await Product.Q.drop_collection(force=True)
    await ProductImage.Q.drop_collection(force=True)


@pytest.mark.asyncio
async def test_aggregation_math_operation(connection):
    max_ = await Product.Q.simple_aggregate(aggregation=Max('cost'))
    assert max_.data == {'cost__max': 4}

    min_ = await Product.Q.simple_aggregate(aggregation=Min('cost'))
    assert min_.data == {'cost__min': 1}

    sum_ = await Product.Q.simple_aggregate(aggregation=Sum('cost'))
    assert sum_.data == {'cost__sum': 10}

    avg_ = await Product.Q.simple_aggregate(aggregation=Avg('cost'))
    assert avg_.data == {'cost__avg': 2.5}

    simple_avg = await Product.Q.aggregate_sum('cost')
    assert simple_avg == 10.0

    simple_max = await Product.Q.aggregate_max('cost')
    assert simple_max == 4

    simple_min = await Product.Q.aggregate_min('cost')
    assert simple_min == 1

    simple_avg = await Product.Q.aggregate_avg('cost')
    assert simple_avg == 2.5


@pytest.mark.asyncio
async def test_aggregation_multiply(connection):
    result_sum = await Product.Q.simple_aggregate(
        aggregation=[Sum('cost'), Sum('quantity')]
    )
    assert result_sum.data == {'cost__sum': 10.0, 'quantity__sum': 10}

    result_max = await Product.Q.simple_aggregate(
        aggregation=[Max('cost'), Max('quantity')]
    )
    assert result_max.data == {'cost__max': 4.0, 'quantity__max': 4}

    result_min = await Product.Q.simple_aggregate(
        aggregation=[Min('cost'), Min('quantity')]
    )
    assert result_min.data == {'cost__min': 1.0, 'quantity__min': 1}

    result_avg = await Product.Q.simple_aggregate(
        aggregation=(Avg('cost'), Avg('quantity'))
    )
    assert result_avg.data == {'cost__avg': 2.5, 'quantity__avg': 2.5}

    result_multiply = await Product.Q.simple_aggregate(
        aggregation=(Avg('cost'), Max('quantity'))
    )
    assert result_multiply.data == {'cost__avg': 2.5, 'quantity__max': 4}

    result_count = await Product.Q.simple_aggregate(aggregation=Count('product_type'))
    assert result_count.data == {'book': {'count': 3}, 'phone': {'count': 1}}

    result_count_agg = await Product.Q.simple_aggregate(
        aggregation=[Count('product_type'), Sum('cost')]
    )
    assert result_count_agg.data == {
        'book': {'cost__sum': 6.0, 'count': 3},
        'phone': {'cost__sum': 4.0, 'count': 1},
    }

    result_sum_and_avg_agg_with_group = await Product.Q.simple_aggregate(
        aggregation=[Avg('cost'), Sum('cost')],
        group_by=['product_type'],
    )
    assert result_sum_and_avg_agg_with_group.data == {
        'phone': {'cost__avg': 4.0, 'cost__sum': 4.0},
        'book': {'cost__avg': 2.0, 'cost__sum': 6.0},
    }

    result_group_by_by_inners = await Product.Q.simple_aggregate(
        group_by=['config.type_id'], aggregation=Count('_id')
    )
    assert result_group_by_by_inners.data == {
        '4': {'count': 1},
        '3': {'count': 1},
        '2': {'count': 1},
        '1': {'count': 1},
    }

    result_sum_and_avg_agg_with_group_many = await Product.Q.simple_aggregate(
        aggregation=[Avg('cost'), Sum('cost')],
        group_by=['product_type', 'quantity'],
    )
    assert result_sum_and_avg_agg_with_group_many.data == {
        'phone|4': {'cost__avg': 4.0, 'cost__sum': 4.0},
        'book|3': {'cost__avg': 3.0, 'cost__sum': 3.0},
        'book|2': {'cost__avg': 2.0, 'cost__sum': 2.0},
        'book|1': {'cost__avg': 1.0, 'cost__sum': 1.0},
    }

    result_agg = await Product.Q.simple_aggregate(
        aggregation=[Avg('cost'), Max('quantity')]
    )
    assert result_agg.data == {'cost__avg': 2.5, 'quantity__max': 4}

    result_not_match_agg = await Product.Q.simple_aggregate(
        Q(title__ne='not_match') & Q(title__startswith='not'),
        aggregation=[Avg('cost'), Max('quantity')],
    )
    assert result_not_match_agg.data == {}


@pytest.mark.asyncio
async def test_aggregate_method(connection):
    aggregate = (
        Product.manager.aggregate()
        .match(title__in=['1', '2', '3'])
        .group(_id={"quant": '$quantity'})
        .skip(1)
        .limit(1)
    )
    match_pipeline = [
        {'$match': {'title': {'$in': ['1', '2', '3']}}},
        {'$group': {'_id': {'quant': '$quantity'}}},
        {'$skip': 1},
        {'$limit': 1},
    ]
    assert aggregate.pipeline == match_pipeline
    result = await aggregate.result()
    assert result == [{'_id': {'quant': 2}}]

    image_aggreagate = (
        ProductImage.manager.aggregate()
        .lookup(
            Product,
            local_field='product_id',
            foreign_field='_id',
            as_='product',
        )
        .unwind("$product")
        .project({'product_title': '$product.title', '_id': 0, 'url': 1})
        .add_fields(product_image_url={'$concat': ['$product_title', ' url: ', '$url']})
    )
    assert image_aggreagate.pipeline == [
        {
            '$lookup': {
                'from': 'product',
                'localField': 'product_id',
                'foreignField': '_id',
                'as': 'product',
            }
        },
        {'$unwind': '$product'},
        {'$project': {'product_title': '$product.title', '_id': 0, 'url': 1}},
        {
            '$addFields': {
                'product_image_url': {'$concat': ['$product_title', ' url: ', '$url']}
            }
        },
    ]
    result = await image_aggreagate.result()
    assert (
        'url' in result[0]
        and 'product_title' in result[0]
        and 'product_image_url' in result[0]
    )


@pytest.mark.asyncio
async def test_logical_match_query_in_aggregate(connection):
    aggregate = (
        Product.manager.aggregate().match(Q(title=1) | Q(title=4)).sort(title=-1)
    )
    assert aggregate.pipeline == [
        {'$match': {'$or': [{'title': '1'}, {'title': '4'}]}},
        {'$sort': {'title': -1}},
    ]
    result = await aggregate.result()

    assert result[0]['title'] == '4'
    assert result[-1]['title'] == '1'


@pytest.mark.asyncio
async def test_geo(connection):
    from pymongo import IndexModel

    class Place(Document):
        name: str
        location: dict
        category: str

        class Config:
            indexes = [IndexModel([('location', '2dsphere')])]

    await Place.ensure_indexes()

    data = [
        {
            'name': "Central Park",
            'location': {'type': "Point", 'coordinates': [-73.97, 40.77]},
            'category': "Parks",
        },
        {
            'name': "Sara D. Roosevelt Park",
            'location': {'type': "Point", 'coordinates': [-73.9928, 40.7193]},
            'category': "Parks",
        },
        {
            'name': "Polo Grounds",
            'location': {'type': "Point", 'coordinates': [-73.9375, 40.8303]},
            'category': "Stadiums",
        },
    ]
    await Place.Q.insert_many(data)

    geo_aggregate = (
        Place.manager.aggregate()
        .geo_near(
            near={'type': 'Point', 'coordinates': [-73.99279, 40.719296]},
            distance_field='dist.calculated',
            max_distance=2,
            query={'category': 'Parks'},
            include_locs='dist.location',
            spherical=True,
        )
        .project({'_id': 0})
    )
    assert geo_aggregate.pipeline == [
        {
            '$geoNear': {
                'distanceField': 'dist.calculated',
                'near': {'type': 'Point', 'coordinates': [-73.99279, 40.719296]},
                'includeLocs': 'dist.location',
                'maxDistance': 2,
                'query': {'category': 'Parks'},
                'spherical': True,
            }
        },
        {'$project': {'_id': 0}},
    ]
    result = await geo_aggregate.result()
    assert result[0] == {
        'name': 'Sara D. Roosevelt Park',
        'location': {'type': 'Point', 'coordinates': [-73.9928, 40.7193]},
        'category': 'Parks',
        'dist': {
            'calculated': 0.9539931676365992,
            'location': {'type': 'Point', 'coordinates': [-73.9928, 40.7193]},
        },
    }
    await Place.Q.drop_collection(force=True)


@pytest.mark.asyncio
async def test_raises_invalid_field(connection):
    with pytest.raises(MotordanticValidationError):
        await Product.Q.simple_aggregate(
            title='not_match', aggregation=[Avg('cost123'), Max('quantityzzzz')]
        )


@pytest.mark.asyncio
async def test_raw_aggregate(connection):
    result_raw_group_by_by_inners = await Product.Q.raw_aggregate(
        data=[
            {
                "$group": {
                    "_id": {'type_id': "$config.type_id"},
                    'count': {f'$sum': 1},
                    'names': {'$push': '$title'},
                }
            },
        ],
    )
    assert result_raw_group_by_by_inners == [
        {'_id': {'type_id': 4}, 'count': 1, 'names': ['4']},
        {'_id': {'type_id': 3}, 'count': 1, 'names': ['3']},
        {'_id': {'type_id': 2}, 'count': 1, 'names': ['2']},
        {'_id': {'type_id': 1}, 'count': 1, 'names': ['1']},
    ]



from typing import Type

from ..query.query import QueryBuilder, BaseQuery

__all__ = ('SyncQuery', 'SyncQueryBuilder')


class SyncQuery(BaseQuery):
    def __getattr__(self, method_name: str) -> 'SyncQuery':
        return SyncQuery(self._builder, method_name)

    def __call__(self, *args, **kwargs):
        method = getattr(self._builder, self.method_name)
        return self._builder.odm_manager._io_loop.run_until_complete(
            method(*args, **kwargs)
        )


class SyncQueryBuilder(QueryBuilder):
    query_class: Type[BaseQuery] = SyncQuery



import pytest
import pytest_asyncio

from motordantic.exceptions import MotordanticValidationError
from motordantic.document import Document


class TestField(str):
    type_ = str
    required = False
    default = None
    validate_always = False
    alias = ''

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, value):
        if not value or value == 'no':
            raise ValueError('invalid value')
        return value


class Validate(Document):
    name: str
    position: int
    config: dict
    sign: int = 1
    type_: str = 'ga'
    test: TestField = TestField('test value')
    array: list = [1, 2]


@pytest_asyncio.fixture(scope='session', autouse=True)
async def drop_ticket_collection(event_loop):
    yield
    await Validate.Q.drop_collection(force=True)


@pytest.mark.asyncio
async def test_validate(connection):
    with pytest.raises(MotordanticValidationError):
        await Validate.Q.find_one(array='invalid')

    with pytest.raises(MotordanticValidationError):
        await Validate.Q.find_one(_id='invalid')

    with pytest.raises(MotordanticValidationError):
        await Validate.Q.find_one(position='invalid')
        await Validate.Q.find_one(config='invalid')

    with pytest.raises(MotordanticValidationError):
        Validate(
            array={'invalid': 'true'},
            name='123',
            position='invalid postion',
            config={},
        )

    with pytest.raises(MotordanticValidationError):
        Validate(array=[], name='123', position=1, config={}, test='no')



import pytest
import re

from bson import Regex

from motordantic.document import Document
from motordantic.query.extra import ExtraQueryMapper


class User(Document):
    id: str
    name: str
    counter: int
    date: str


def test_in_extra_param():
    with pytest.raises(TypeError):
        ExtraQueryMapper(User, 'name').query(['in'], (1, 3))
    extra = ExtraQueryMapper(User, 'name').query(['in'], [1, 3])
    value = {'name': {'$in': ['1', '3']}}
    assert extra == value


def test_nin_extra_param():
    with pytest.raises(TypeError):
        ExtraQueryMapper(User, 'name').query(['nin'], (1, 3))
    extra = ExtraQueryMapper(User, 'name').query(['nin'], [1, 3])
    value = {'name': {'$nin': ['1', '3']}}
    assert extra == value


def test_ne_extra_param():
    extra = ExtraQueryMapper(User, 'name').query(['ne'], 'test')
    value = {'name': {"$ne": 'test'}}
    assert extra == value


def test_regex_extra_param():
    extra = ExtraQueryMapper(User, 'name').query(['regex'], 'test')
    value = {'name': {'$regex': Regex.from_native(re.compile('test'))}}
    assert extra == value


def test_regex_ne_extra_param():
    extra = ExtraQueryMapper(User, 'name').query(['regex_ne'], 'test')
    value = {'name': {"$not": Regex.from_native(re.compile('test'))}}
    assert extra == value


def test_startswith_extra_param():
    extra = ExtraQueryMapper(User, 'name').query(['startswith'], 'test')
    value = {'name': {'$regex': Regex.from_native(re.compile('^test'))}}
    assert extra == value


def test_endswith_extra_param():
    extra = ExtraQueryMapper(User, 'name').query(['endswith'], 'test')
    value = {'name': {'$regex': Regex.from_native(re.compile('test$'))}}
    assert extra == value


def test_not_endswith_extra_param():
    extra = ExtraQueryMapper(User, 'name').query(['not_endswith'], 'test')
    value = {'name': {"$not": Regex.from_native(re.compile('test$'))}}
    assert extra == value


def test_not_startswith_extra_param():
    extra = ExtraQueryMapper(User, 'name').query(['not_startswith'], 'test')
    value = {'name': {"$not": Regex.from_native(re.compile('^test'))}}
    assert extra == value


def test_range_extra_param():
    with pytest.raises(ValueError):
        ExtraQueryMapper(User, 'date').query(['range'], 'test')
    extra = ExtraQueryMapper(User, 'date').query(
        ['range'], ['2020-01-01', '2020-03-01']
    )
    value = {'date': {"$gte": '2020-01-01', "$lte": '2020-03-01'}}
    assert extra == value


def test_lts_gts_params():
    extra = ExtraQueryMapper(User, 'date').query(['lt'], '2020-01-01')
    assert extra == {'date': {"$lt": '2020-01-01'}}

    extra = ExtraQueryMapper(User, 'date').query(['gt'], '2020-01-01')
    assert extra == {'date': {"$gt": '2020-01-01'}}

    extra = ExtraQueryMapper(User, 'date').query(['gte'], '2020-01-01')
    assert extra == {'date': {"$gte": '2020-01-01'}}

    extra = ExtraQueryMapper(User, 'date').query(['lte'], '2020-01-01')
    assert extra == {'date': {"$lte": '2020-01-01'}}


def test_inc_params():
    with pytest.raises(ValueError):
        ExtraQueryMapper(User, 'counter').query(['inc'], '2313123131')

    extra = ExtraQueryMapper(User, 'counter').query(['inc'], 23)
    assert extra == {'$inc': {'counter': 23}}


def test_exists_params():
    with pytest.raises(TypeError):
        ExtraQueryMapper(User, 'counter').query(['exists'], '2313123131')

    extra = ExtraQueryMapper(User, 'counter').query(['exists'], False)
    assert extra == {'counter': {'$exists': False}}



import asyncio
from typing import TYPE_CHECKING, List

from .types import RelationTypes

__all__ = ('RelationManager',)

if TYPE_CHECKING:
    from .document import Document
    from .typing import DictStrList, DocumentType


class RelationManager(object):
    """relation manager for get and set data from to to models instances"""

    __slots__ = ('document_class', 'relation_fields')

    def __init__(self, document_class: 'Document'):
        self.document_class = document_class
        self.relation_fields = self._get_relation_fields(document_class)

    @classmethod
    def _get_relation_fields(cls, document_class: 'Document') -> dict:
        return document_class.__db_refs__ or {}

    def _relation_data_setter(self, document: 'Document', data: dict) -> 'Document':
        """data setter

        Args:
            document (Document): mongo model insatance
            data (dict): relations objects map

        Returns:
            Document: updated mongo model
        """
        for field, relation_info in self.relation_fields.items():
            relation_attr = getattr(document, field)
            if not relation_attr:
                continue
            if relation_info.relation_type == RelationTypes.ARRAY:
                relation_value = []
                for rel in relation_attr:
                    v = data[field].get(str(rel.db_ref.id))
                    if v:
                        relation_value.append(v)
            else:
                relation_value = data[field].get(str(relation_attr.db_ref.id))
            setattr(document, field, relation_value)
        return document

    async def _get_relation_objects_by_model_class(
        self, field: str, document_class: 'DocumentType', ids: list
    ) -> dict:
        result = await document_class.Q.find(_id__in=ids, with_relations_objects=True)
        return {field: {str(o._id): o for o in result}}

    async def get_relation_objects(self, pre_relation: 'DictStrList') -> dict:
        futures = []
        for field, ids in pre_relation.items():
            document_class = self.relation_fields[field].document_class
            futures.append(
                asyncio.ensure_future(
                    self._get_relation_objects_by_model_class(
                        field, document_class, ids
                    )
                )
            )

        relation_objects = {}
        for _, field_relation_result in enumerate(
            asyncio.as_completed(futures), start=1
        ):
            relation_objects.update(await field_relation_result)
        return relation_objects

    def _get_pre_relation(self, document_instances: List['Document']) -> 'DictStrList':
        pre_relation: 'DictStrList' = {field: [] for field in self.relation_fields}
        for document_instance in document_instances:
            for field in self.relation_fields:
                attr = getattr(document_instance, field)
                if isinstance(attr, list):
                    ids = tuple(row.db_ref.id for row in attr)
                else:
                    ids = (attr.db_ref.id,) if attr else tuple()
                if ids:
                    pre_relation[field].extend(ids)
        return {f: list(set(items)) for f, items in pre_relation.items()}

    async def map_relation_for_single(
        self, document_instance: 'Document'
    ) -> 'Document':
        """map relation data to mongo model instanc

        Args:
            document_instances (List[Document]): list of instances from Document

        Returns:
            Document: mapped mongo model
        """
        pre_relation: 'DictStrList' = self._get_pre_relation([document_instance])
        relation_objects = await self.get_relation_objects(pre_relation)
        return self._relation_data_setter(document_instance, relation_objects)

    async def map_relation_for_array(self, result: List) -> List['Document']:
        """map relations data for _find method list result

        Args:
            result (List): _find query result converted to list

        Returns:
            List: mapped list
        """
        pre_relation: 'DictStrList' = self._get_pre_relation(result)
        relation_objects = await self.get_relation_objects(pre_relation)
        generated_result = [
            self._relation_data_setter(r, relation_objects) for r in result
        ]
        return generated_result



from json import dumps
from typing import Generator, List, Union, Any, Tuple, List, TYPE_CHECKING, Union

if TYPE_CHECKING:
    from ..document import Document


class AggregateResult(object):
    __slots__ = ('_native_result', '_document_class')

    def __init__(self, native_result: list, document_class: 'Document'):
        self._native_result = native_result
        self._document_class = document_class

    def __iter__(self):
        for obj in self.native_result:
            yield obj

    def __next__(self):
        return next(self.__iter__())

    def __len__(self):
        return len(self.native_result)

    def __contains__(self, value):
        return value in self.native_result

    def __lt__(self, other):
        return self.native_result < self.__cast(other)

    def __le__(self, other):
        return self.native_result <= self.__cast(other)

    def __eq__(self, other):
        return self.native_result == self.__cast(other)

    def __gt__(self, other):
        return self.native_result > self.__cast(other)

    def __ge__(self, other):
        return self.native_result >= self.__cast(other)

    def __cast(self, other):
        return other.native_result if isinstance(other, AggregateResult) else other

    def __getitem__(self, idx):
        if isinstance(idx, slice):
            return self.__class__(self.native_result[idx])  # type: ignore
        else:
            return self.native_result[idx]

    @property
    def native_result(self) -> list:
        return self._native_result

    @property
    def document_class(self) -> 'Document':
        return self._document_class


class SimpleAggregateResult(object):
    __slots__ = ('_data', 'document_class')

    def __init__(
        self,
        document_class: 'Document',
        data: dict,
    ):
        self._data = data
        self.document_class = document_class

    def json(self) -> str:
        return dumps(self._data)

    @property
    def data(self) -> dict:
        return self._data


class FindResult(object):
    __slots__ = ('_data', 'document_class')

    def __init__(
        self,
        document_class: 'Document',
        data: list,
    ):
        self._data = data
        self.document_class = document_class

    # @handle_and_convert_connection_errors
    def __iter__(self):
        for obj in self._data:
            yield obj

    def __next__(self):
        return next(self.__iter__())

    @property
    def data(self) -> List:
        return [obj.data for obj in self.__iter__()]

    @property
    def generator(self) -> Generator:
        return self.__iter__()

    @property
    def data_generator(self) -> Generator:
        for obj in self.__iter__():
            yield obj.data

    @property
    def list(self) -> List:
        return list(self.__iter__())

    def json(self) -> str:
        return dumps(self.data)

    def first(self) -> Any:
        return next(self.__iter__())

    def serialize(
        self, fields: Union[Tuple, List], to_list: bool = True
    ) -> Union[Tuple, List]:
        return (
            [obj.serialize(fields) for obj in self.__iter__()]
            if to_list
            else tuple(obj.serialize(fields) for obj in self.__iter__())
        )

    def serialize_generator(self, fields: Union[Tuple, List]) -> Generator:
        for obj in self.__iter__():
            yield obj.serialize(fields)

    def serialize_json(self, fields: Union[Tuple, List]) -> str:
        return dumps(self.serialize(fields))



from typing import (
    AsyncIterable,
    Union,
    List,
    Dict,
    Optional,
    Any,
    Tuple,
    TYPE_CHECKING,
    Iterable,
)

from bson import ObjectId, decode as bson_decode
from bson.raw_bson import RawBSONDocument
from pymongo import ReturnDocument, IndexModel
from pymongo.collection import WriteConcern
from motor.core import AgnosticClientSession as ClientSession

from .query import generate_basic_query, Q, QCombination
from .result import FindResult, SimpleAggregateResult
from .extra import group_by_aggregate_generation, generate_name_field

from ..aggregate.expressions import Sum, Max, Min, Avg
from ..exceptions import (
    MotordanticInvalidArgsParams,
    MotordanticValidationError,
    MotordanticIndexError,
    DoesNotExist,
)
from ..validation import sort_validation


__all__ = ('Builder',)

if TYPE_CHECKING:
    from ..manager import ODMManager
    from ..typing import DictStrAny
    from ..document import Document


class Builder(object):
    __slots__ = ('odm_manager',)

    def __init__(self, odm_manager: 'ODMManager'):
        self.odm_manager: 'ODMManager' = odm_manager

    def _validate_query_data(self, query: Dict) -> 'DictStrAny':
        """main validation method

        Args:
            query (Dict): basic query

        Returns:
            Dict: parsed query
        """
        query_params: dict = generate_basic_query(self.odm_manager, query)

        return query_params

    def _check_query_args(
        self,
        logical_query: Union[
            List[Any], Dict[Any, Any], str, Q, QCombination, None
        ] = None,
    ) -> 'DictStrAny':
        """check if query = Query obj or QueryCombination

        Args:
            logical_query (Union[ List[Any], Dict[Any, Any], str, Q, QCombination ], optional): Query | QueryCombination. Defaults to None.

        Raises:
            InvalidArgsParams: if not Query | QueryCombination

        Returns:
            Dict: generated query dict
        """
        if not isinstance(logical_query, (Q, QCombination)):
            raise MotordanticInvalidArgsParams()
        return logical_query.to_query(self)  # type: ignore

    async def _make_query(
        self,
        method_name: str,
        query_params: Union[List, Dict, str, Q, QCombination],
        set_values: Optional[Dict] = None,
        session: Optional[ClientSession] = None,
        logical: bool = False,
        write_concern: Optional[WriteConcern] = None,
        **kwargs,
    ) -> Any:
        """main query function

        Args:
            method_name (str): query method like find, find_one and other
            query_params (Union[List, Dict, str, Query, LogicalCombination]): query params: dict or Query or LogicalCombination
            set_values (Optional[Dict], optional): for updated method. Defaults to None.
            session (Optional[ClientSession], optional): motor session. Defaults to None.
            logical (bool, optional): if logical. Defaults to False.

        Returns:
            Any: query result
        """
        if logical:
            query_params = self._check_query_args(query_params)
        elif isinstance(query_params, dict):
            query_params = self._validate_query_data(query_params)
        if write_concern:
            collection = self.odm_manager.collection.with_options(  # type: ignore
                write_concern=write_concern
            )
        else:
            collection = self.odm_manager.collection
        # print(query_params)
        method = getattr(collection, method_name)
        query: tuple = (query_params,)
        if session:
            kwargs['session'] = session
        if set_values:
            query = (query_params, set_values)
        if kwargs:
            return await method(*query, **kwargs)
        return await method(*query)

    async def get(self, session: Optional[ClientSession] = None, **query) -> 'Document':
        obj = await self.find_one(session=session, **query)
        if not obj:
            raise DoesNotExist(self.odm_manager.document.__name__)
        return obj

    async def count(
        self,
        logical_query: Union[Q, QCombination, None] = None,
        session: Optional[ClientSession] = None,
        **query,
    ) -> int:
        """count query

        Args:
            logical_query (Union[Q, QCombination, None], optional): Query | QueryCombination. Defaults to None.
            session (Optional[ClientSession], optional): motor session. Defaults to None.

        Returns:
            int: count of documents
        """
        if getattr(self.odm_manager.collection, 'count_documents'):
            return await self._make_query(
                'count_documents',
                logical_query or query,
                session=session,
                logical=bool(logical_query),
            )
        return await self._make_query(
            'count',
            logical_query or query,
            session=session,
            logical=bool(logical_query),
        )

    async def count_documents(
        self,
        logical_query: Union[Q, QCombination, None] = None,
        session: Optional[ClientSession] = None,
        **query,
    ) -> int:
        """count_documents query

        Args:
            logical_query (Union[Query, LogicalCombination, None], optional): Query | QueryCombination. Defaults to None.
            session (Optional[ClientSession], optional): motor session. Defaults to None.

        Returns:
            int: count of documents
        """
        return await self.count(logical_query, session, **query)

    async def insert_one(
        self, session: Optional[ClientSession] = None, **query
    ) -> ObjectId:
        """insert one document

        Args:
            session (Optional[ClientSession], optional): motor session. Defaults to None.

        Returns:
            ObjectId: created document _id
        """
        obj = self.odm_manager.document.parse_obj(query)
        data = await self._make_query(
            'insert_one',
            obj._mongo_query_data,
            session=session,
        )
        return data.inserted_id

    async def insert_many(
        self,
        data: List,
        session: Optional[ClientSession] = None,
        ordered: bool = True,
        bypass_document_validation: bool = False,
        write_concern: Optional[WriteConcern] = None,
    ) -> int:
        """insert many documents

        Args:
            data (List): List of dict or Documents
            session (Optional[ClientSession], optional): motor session. Defaults to None.

        Returns:
            int: count inserted ids
        """
        parse_obj = self.odm_manager.document.parse_obj
        query = [
            parse_obj(obj)._mongo_query_data
            if isinstance(obj, dict)
            else obj._mongo_query_data
            for obj in data
        ]
        r = await self._make_query(
            'insert_many',
            query,
            session=session,
            ordered=ordered,
            bypass_document_validation=bypass_document_validation,
            write_concern=write_concern,
        )
        return len(r.inserted_ids)

    async def delete_one(
        self,
        logical_query: Union[Q, QCombination, None] = None,
        session: Optional[ClientSession] = None,
        **query,
    ) -> int:
        """delete one document

        Args:
            logical_query (Union[Q, QCombination, None], optional): Query|QueryCombination. Defaults to None.
            session (Optional[ClientSession], optional): motor session. Defaults to None.

        Returns:
            int: deleted documents count
        """
        r = await self._make_query(
            'delete_one',
            logical_query or query,
            session=session,
            logical=bool(logical_query),
        )
        return r.deleted_count

    async def delete_many(
        self,
        logical_query: Union[Q, QCombination, None] = None,
        session: Optional[ClientSession] = None,
        **query,
    ) -> int:
        """delete many document

        Args:
            logical_query (Union[Q, QCombination, None], optional): Query|QueryCombination. Defaults to None.
            session (Optional[ClientSession], optional): motor session. Defaults to None.

        Returns:
            int: deleted documents count
        """
        r = await self._make_query(
            'delete_many',
            logical_query or query,
            session=session,
            logical=bool(logical_query),
        )
        return r.deleted_count

    async def distinct(
        self, field: str, session: Optional[ClientSession] = None, **query
    ) -> list:
        """wrapper for pymongo distinct

        Args:
            field (str): distinct field
            session (Optional[ClientSession], optional): motor session. Defaults to None.

        Returns:
            list: list of distinct values
        """
        query = self._validate_query_data(query)
        method = getattr(self.odm_manager.collection, 'distinct')
        return await method(key=field, filter=query, session=session)

    async def find_one(
        self,
        logical_query: Union[Q, QCombination, None] = None,
        sort_fields: Optional[Union[Tuple, List]] = None,
        session: Optional[ClientSession] = None,
        sort: Optional[int] = None,
        with_relations_objects: bool = False,
        **query,
    ) -> Optional['Document']:
        """find one document

        Args:
            logical_query (Union[Q, QCombination, None], optional): Query | LogicalCombination. Defaults to None.
            session (Optional[ClientSession], optional): motor session. Defaults to None.
            sort_fields (Optional[Union[Tuple, List]], optional): iterable from sort fielda. Defaults to None.
            sort (Optional[int], optional): sort value -1 or 1. Defaults to None.

        Returns:
            Optional[Document]: Document instance or None
        """
        sort, sort_fields = sort_validation(sort, sort_fields)
        data = await self._make_query(
            'find_one',
            logical_query or query,
            logical=bool(logical_query),
            sort=[(field, sort or 1) for field in sort_fields] if sort_fields else None,
            session=session,
        )
        if data is not None:
            obj = self.odm_manager.document.from_bson(data)
            if with_relations_objects and self.odm_manager.relation_manager:
                obj = await self.odm_manager.relation_manager.map_relation_for_single(
                    obj
                )
            return obj
        return None

    async def _find(
        self,
        logical_query: Union[Q, QCombination, None] = None,
        skip_rows: Optional[int] = None,
        limit_rows: Optional[int] = None,
        session: Optional[ClientSession] = None,
        sort_fields: Optional[Union[Tuple, List]] = None,
        sort: Optional[int] = None,
        **query,
    ) -> AsyncIterable:
        sort, sort_fields_parsed = sort_validation(sort, sort_fields)

        async def context():
            if bool(logical_query):
                query_params = self._check_query_args(logical_query)
            else:
                query_params = self._validate_query_data(query)
            find_cursor_method = getattr(self.odm_manager.collection, 'find')
            cursor = find_cursor_method(query_params, session=session)
            if skip_rows is not None:
                cursor = cursor.skip(skip_rows)
            if limit_rows:
                cursor = cursor.limit(limit_rows)
            if sort:
                cursor.sort([(field, sort or 1) for field in sort_fields_parsed])
            async for doc in cursor:
                yield self.odm_manager.document.from_bson(doc)

        return context()

    async def find(
        self,
        logical_query: Union[Q, QCombination, None] = None,
        skip_rows: Optional[int] = None,
        limit_rows: Optional[int] = None,
        session: Optional[ClientSession] = None,
        sort_fields: Optional[Union[Tuple, List]] = None,
        sort: Optional[int] = None,
        with_relations_objects: bool = False,
        **query,
    ) -> FindResult:
        """find method

        Args:
            logical_query (Union[Query, LogicalCombination, None], optional): Query|LogicalCombunation. Defaults to None.
            skip_rows (Optional[int], optional): skip rows for pagination. Defaults to None.
            limit_rows (Optional[int], optional): limit rows. Defaults to None.
            session (Optional[ClientSession], optional): pymongo session. Defaults to None.
            sort_fields (Optional[Union[Tuple, List]], optional): iterable from sort fielda. Defaults to None.
            sort (Optional[int], optional): sort value -1 or 1. Defaults to None.

        Returns:
            FindResult: Motordantic FindResult
        """
        result = await self._find(
            logical_query,
            skip_rows,
            limit_rows,
            session,
            sort_fields,
            sort,
            **query,
        )
        data = [doc async for doc in result]
        if with_relations_objects and self.odm_manager.relation_manager:
            data = await self.odm_manager.relation_manager.map_relation_for_array(data)
        return FindResult(self.odm_manager.document, data)

    def _prepare_update_data(self, **fields) -> tuple:
        """prepare and validate query data for update queries"""

        if not any("__set" in f for f in fields):
            raise MotordanticValidationError("not fields for updating!")
        query_params = {}
        set_values = {}
        for name, value in fields.items():
            if name.endswith('__set'):
                name = name.replace('__set', '')
                data = self._validate_query_data({name: value})
                set_values.update(data)
            else:
                query_params.update({name: value})
        return query_params, set_values

    async def _update(
        self,
        method: str,
        query: Dict,
        upsert: bool = True,
        session: Optional[ClientSession] = None,
    ) -> int:
        """innert method for update

        Args:
            method (str): one of update_many or update_one
            query (Dict): update query
            upsert (bool, optional): upsert option. Defaults to True.
            session (Optional[ClientSession], optional): motor session. Defaults to None.

        Returns:
            int: updated documents count
        """
        query, set_values = self._prepare_update_data(**query)
        r = await self._make_query(
            method, query, {'$set': set_values}, upsert=upsert, session=session
        )
        return r.modified_count

    async def update_one(
        self, upsert: bool = False, session: Optional[ClientSession] = None, **query
    ) -> int:
        """update one document

        Args:
            upsert (bool, optional): pymongo upsert. Defaults to False.
            session (Optional[ClientSession], optional): motor session. Defaults to None.

        Returns:
            int: updated documents count
        """
        return await self._update('update_one', query, upsert=upsert, session=session)

    async def update_many(
        self, upsert: bool = False, session: Optional[ClientSession] = None, **query
    ) -> int:
        """update many document

        Args:
            upsert (bool, optional): pymongo upsert. Defaults to False.
            session (Optional[ClientSession], optional): motor session. Defaults to None.

        Returns:
            int: updated documents count
        """
        return await self._update('update_many', query, upsert=upsert, session=session)

    async def _find_with_replacement_or_with_update(
        self,
        operation: str,
        projection_fields: Optional[list] = None,
        sort_fields: Optional[Union[Tuple, List]] = None,
        sort: Optional[int] = None,
        upsert: bool = False,
        session: Optional[ClientSession] = None,
        **query,
    ) -> Union[Dict, 'Document', None]:
        """base method for find_with_<operation>

        Args:
            operation (str): operation name
            projection_fields (Optional[list], optional): prejection. Defaults to None.
            sort_fields (Optional[Union[Tuple, List]], optional): sort fields. Defaults to None.
            sort (Optional[int], optional): -1 or 1. Defaults to None.
            upsert (bool, optional): True/False. Defaults to False.
            session (Optional[ClientSession], optional): motor session. Defaults to None.

        Returns:
            Union[Dict, 'Document']: Document or Dict or None
        """
        filter_, set_values = self._prepare_update_data(**query)
        return_document = ReturnDocument.AFTER
        replacement = query.pop('replacement', None)

        projection = {f: True for f in projection_fields} if projection_fields else None
        extra_params = {
            'return_document': return_document,
            'projection': projection,
            'upsert': upsert,
            'session': session,
        }
        if sort_fields:
            extra_params['sort'] = [(field, sort or 1) for field in sort_fields]

        if replacement:
            extra_params['replacement'] = replacement

        data = await self._make_query(
            operation, filter_, set_values={'$set': set_values}, **extra_params
        )
        if projection:
            return {
                field: value for field, value in data.items() if field in projection
            }
        if data:
            return self.odm_manager.document.from_bson(data)
        return None

    async def find_one_and_update(
        self,
        projection_fields: Optional[list] = None,
        sort_fields: Optional[Union[Tuple, List]] = None,
        sort: Optional[int] = None,
        upsert: bool = False,
        session: Optional[ClientSession] = None,
        **query,
    ) -> Union[Dict, 'Document', None]:
        """find one and update

        Args:
            operation (str): operation name
            projection_fields (Optional[list], optional): prejection. Defaults to None.
            sort_fields (Optional[Union[Tuple, List]], optional): sort fields. Defaults to None.
            sort (Optional[int], optional): -1 or 1. Defaults to None.
            upsert (bool, optional): True/False. Defaults to False.
            session (Optional[ClientSession], optional): motor session. Defaults to None.

        Returns:
            Union[Dict, 'Document']: Document or Dict or None
        """
        return await self._find_with_replacement_or_with_update(
            'find_one_and_update',
            projection_fields=projection_fields,
            sort_fields=[(field, sort or 1) for field in sort_fields]
            if sort_fields
            else None,
            sort=sort,
            upsert=upsert,
            session=session,
            **query,
        )

    async def find_and_replace(
        self,
        replacement: Union[dict, Any],
        projection_fields: Optional[list] = None,
        sort_fields: Optional[Union[Tuple, List]] = None,
        sort: Optional[int] = None,
        upsert: bool = False,
        session: Optional[ClientSession] = None,
        **query,
    ) -> Union[Dict, 'Document', None]:
        """find one and replace

        Args:
            operation (str): operation name
            projection_fields (Optional[list], optional): prejection. Defaults to None.
            sort_fields (Optional[Union[Tuple, List]], optional): sort fields. Defaults to None.
            sort (Optional[int], optional): -1 or 1. Defaults to None.
            upsert (bool, optional): True/False. Defaults to False.
            session (Optional[ClientSession], optional): motor session. Defaults to None.

        Returns:
            Union[Dict, 'Document']: Document or Dict or None
        """
        if not isinstance(replacement, dict):
            replacement = replacement.query_data
        return await self._find_with_replacement_or_with_update(
            'find_and_replace',
            projection_fields=projection_fields,
            sort_fields=[(field, sort) for field in sort_fields]
            if sort_fields
            else None,
            sort=sort,
            upsert=upsert,
            session=session,
            replacement=replacement,
            **query,
        )

    async def _motor_aggreggate_call(
        self, data: list, session: Optional[ClientSession]
    ) -> AsyncIterable:
        async def context():
            aggregate_cursor = getattr(self.odm_manager.collection, 'aggregate')

            async for row in aggregate_cursor(data, session=session):
                yield row

        return context()

    def from_bson(self, row: RawBSONDocument) -> dict:
        return bson_decode(row.raw)

    async def raw_aggregate(
        self, data: List[Dict[Any, Any]], session: Optional[ClientSession] = None
    ) -> list:
        """raw aggregation query

        Args:
            data (List[Dict[Any, Any]]): aggregation query
            session (Optional[ClientSession], optional): motor session. Defaults to None.

        Returns:
            list: aggregation result
        """
        result = await self._motor_aggreggate_call(data, session)
        return [self.from_bson(row) async for row in result]

    async def _aggregate(self, *args, **query) -> SimpleAggregateResult:
        """main aggregate method

        Raises:
            MongoValidationError: miss aggregation or group_by

        Returns:
            dict: aggregation result
        """
        session = query.pop('session', None)
        aggregation = query.pop('aggregation', None)
        group_by = query.pop('group_by', None)
        if not aggregation and not group_by:
            raise MotordanticIndexError('miss aggregation or group_by')
        if isinstance(aggregation, Iterable):
            aggregate_query = {}
            for agg in aggregation:
                aggregate_query.update(agg._aggregate_query(self.odm_manager.document))
        elif aggregation is not None:
            aggregate_query = aggregation._aggregate_query(self.odm_manager.document)
        else:
            aggregate_query = {}
        if group_by:
            group_by = group_by_aggregate_generation(group_by)
            aggregate_query.pop('_id', None)
            group_params = {"$group": {"_id": group_by, **aggregate_query}}
        else:
            group_params = {
                "$group": {"_id": None, **aggregate_query}
                if '_id' not in aggregate_query
                else aggregate_query
            }
        data = [
            {
                "$match": self._validate_query_data(query)
                if not args
                else self._check_query_args(*args)
            },
            group_params,
        ]

        async_result = await self._motor_aggreggate_call(data, session)
        result = [self.from_bson(row) async for row in async_result]
        if not result:
            return SimpleAggregateResult(self.odm_manager.document, {})
        result_data = {}
        for r in result:
            name = generate_name_field(r.pop('_id'))
            result_data.update({name: r} if name else r)
        return SimpleAggregateResult(self.odm_manager.document, result_data)

    async def simple_aggregate(self, *args, **kwargs) -> SimpleAggregateResult:
        return await self._aggregate(*args, **kwargs)

    async def aggregate_sum(self, agg_field: str, **query) -> Union[int, float]:
        result_field = self.odm_manager.document.__mapping_query_fields__[agg_field]
        result = await self._aggregate(aggregation=Sum(agg_field), **query)
        return result.data.get(f'{result_field}__sum', 0)

    async def aggregate_max(self, agg_field: str, **query) -> Union[int, float]:
        result = await self._aggregate(aggregation=Max(agg_field), **query)
        result_field = self.odm_manager.document.__mapping_query_fields__[agg_field]
        return result.data.get(f'{result_field}__max', 0)

    async def aggregate_min(self, agg_field: str, **query) -> Union[int, float]:
        result = await self._aggregate(aggregation=Min(agg_field), **query)
        result_field = self.odm_manager.document.__mapping_query_fields__[agg_field]
        return result.data.get(f'{result_field}__min', 0)

    async def aggregate_avg(self, agg_field: str, **query) -> Union[int, float]:
        result = await self._aggregate(aggregation=Avg(agg_field), **query)
        result_field = self.odm_manager.document.__mapping_query_fields__[agg_field]
        return result.data.get(f'{result_field}__avg', 0)

    def _validate_raw_query(
        self, method_name: str, raw_query: Union[Dict, List[Dict], Tuple[Dict]]
    ) -> tuple:
        if (
            'insert' in method_name
            or 'replace' in method_name
            or 'update' in method_name
        ):
            if isinstance(raw_query, list):
                raw_query = list(map(self._validate_query_data, raw_query))
            elif isinstance(raw_query, dict):
                raw_query = self._validate_query_data(raw_query)
            else:
                params = [
                    query[key] if '$' in key else query
                    for query in raw_query
                    for key in query.keys()
                ]
                map(self._validate_query_data, params)
        parsed_query = raw_query if isinstance(raw_query, tuple) else (raw_query,)
        return parsed_query

    async def raw_query(
        self,
        method_name: str,
        raw_query: Union[Dict, List[Dict], Tuple[Dict]],
        session: Optional[ClientSession] = None,
    ) -> Any:
        """pymongo raw query

        Args:
            method_name (str): pymongo method, like insert_one
            raw_query (Union[Dict, List[Dict], Tuple[Dict]]): query data
            session (Optional[ClientSession], optional): motor session. Defaults to None.

        Raises:
            MongoValidationError: raise if invalid data

        Returns:
            Any: pymongo query result
        """
        parsed_query = self._validate_raw_query(method_name, raw_query)
        try:
            query = getattr(self.odm_manager.collection, method_name)
            return await query(*parsed_query, session=session)
        except AttributeError:
            raise MotordanticValidationError('invalid method name')

    async def list_indexes(self, session: Optional[ClientSession] = None) -> dict:
        """get indexes for this collection

        Returns:
            dict: indexes result
        """
        list_indexes_cursor = getattr(self.odm_manager.collection, 'list_indexes')
        query = list_indexes_cursor(session=session)
        index_list = await query.to_list(None)
        return_data = {}
        for index in index_list:
            dict_index = dict(index)
            data = {dict_index['name']: {'key': dict(dict_index['key'])}}
            return_data.update(data)
        return return_data

    async def create_indexes(
        self,
        indexes: List[IndexModel],
        session: Optional[ClientSession] = None,
    ) -> List[str]:
        create_indexes_cursor = getattr(self.odm_manager.collection, 'create_index')
        result = []
        for index in indexes:
            index_value = list(index.document['key'].items())
            res = await create_indexes_cursor(
                index_value,
                session=session,
                background=True,
            )
            result.append(res)
        return result

    async def drop_index(
        self, index_name: str, session: Optional[ClientSession] = None
    ) -> str:
        indexes = await self.list_indexes(session)
        if index_name in indexes:
            await self._make_query('drop_index', index_name, session=session)
            return f'{index_name} dropped.'
        raise MotordanticIndexError(f'invalid index name - {index_name}')

    async def drop_collection(self, force: bool = False) -> bool:
        """drop collection

        Args:
            force (bool, optional): if u wanna force drop. Defaults to False.

        Returns:
            bool: result message
        """
        if force:
            await self.odm_manager.collection.drop()  # type: ignore
            return True
        value = input(
            f'Are u sure for drop this collection - {self.odm_manager.document.__name__.lower()} (y, n)'  # type: ignore
        )
        if value.lower() == 'y':
            await self.odm_manager.collection.drop()  # type: ignore
            return True
        return False



import copy
from typing import TYPE_CHECKING, Union, Dict, Type, Any

from .extra import ExtraQueryMapper

from ..validation import validate_field_value, validate_object_id


__all__ = (
    "Q",
    "QCombination",
)


if TYPE_CHECKING:
    from ..manager import ODMManager
    from ..typing import DictStrAny
    from .builder import Builder


class BaseQuery(object):
    __slots__ = ('_builder', 'method_name')

    def __init__(self, builder: 'Builder', method_name: str):
        self._builder = builder
        self.method_name = method_name

    def __call__(self, *args, **kwargs):
        method = getattr(self._builder, self.method_name)
        return method(*args, **kwargs)


class Query(BaseQuery):
    def __getattr__(self, method_name: str) -> 'Query':
        return Query(self._builder, method_name)


class QueryBuilder(object):
    query_class: Type[BaseQuery] = Query

    __slots__ = ('builder', '_valid_methods')

    def __init__(self, builder: 'Builder'):
        self.builder = builder
        self._valid_methods = {
            attr_name: attr_name
            for attr_name in dir(self.builder)
            if attr_name != 'odm_manager' and not attr_name.endswith('__')
        }

    def __getattr__(self, method_name: str) -> Any:
        if method_name in self._valid_methods:
            return self.query_class(self.builder, method_name)
        raise AttributeError(f'invalid Q attr query: {method_name}')

    def _validate_query_data(self, query: Dict) -> 'DictStrAny':
        return self.builder._validate_query_data(query)

    def _check_query_args(self, *args, **kwargs):
        return self.builder._check_query_args(*args, **kwargs)

    def _validate_raw_query(self, *args, **kwargs):
        return self.builder._validate_raw_query(*args, **kwargs)

    def __call__(self, method_name, *args, **method_kwargs):
        return getattr(self, method_name)(*args, **method_kwargs)


def _validate_query_data(builder: 'Builder', query: dict) -> dict:
    return builder._validate_query_data(query)


class QNodeVisitor(object):
    """Base visitor class for visiting Q-object nodes in a query tree."""

    def prepare_combination(
        self, combination: 'QCombination'
    ) -> Union['QCombination', dict]:
        """Called by QCombination objects."""
        return combination

    def visit_query(self, query: 'Q') -> Union['Q', dict]:
        """Called by (New)Q objects."""
        return query


class SimplificationVisitor(QNodeVisitor):
    def __init__(self, builder: 'Builder'):
        self.builder = builder

    def prepare_combination(
        self, combination: 'QCombination'
    ) -> Union['QCombination', dict]:
        if combination.operation == combination.AND:
            # The simplification only applies to 'simple' queries
            if all(isinstance(node, Q) for node in combination.children):
                queries = [n.query for n in combination.children]
                query = self._query_conjunction(queries)
                return {"$and": query}

        return combination

    def _query_conjunction(self, queries):
        """Merges query dicts - effectively &ing them together."""
        combined_query = []
        for query in queries:
            query = _validate_query_data(self.builder, query)
            combined_query.append(copy.deepcopy(query))
        return combined_query


class QCompilerVisitor(QNodeVisitor):
    """Compiles the nodes in a query tree to a PyMongo-compatible query
    dictionary.
    """

    def __init__(self, builder: 'Builder'):
        self.builder = builder

    def prepare_combination(
        self, combination: 'QCombination'
    ) -> Union['QCombination', dict]:
        operator = "$and"
        if combination.operation == combination.OR:
            operator = "$or"
        return {operator: combination.children}

    def visit_query(self, query: 'Q') -> Union['Q', dict]:
        data = _validate_query_data(self.builder, query.query)
        return data


class QNode(object):
    """Base class for nodes in query trees."""

    AND = 0
    OR = 1

    def to_query(self, builder: 'Builder') -> dict:
        query = self.accept(SimplificationVisitor(builder))
        if not isinstance(query, dict):
            query = query.accept(QCompilerVisitor(builder))
        return query

    def accept(self, visitor):
        raise NotImplementedError

    def _combine(self, other, operation):
        """Combine this node with another node into a QCombination
        object.
        """
        # If the other Q() is empty, ignore it and just use `self`.
        if getattr(other, "empty", True):
            return self

        # Or if this Q is empty, ignore it and just use `other`.
        if self.empty:
            return other

        return QCombination(operation, [self, other])

    @property
    def empty(self):
        return False

    def __or__(self, other):
        return self._combine(other, self.OR)

    def __and__(self, other):
        return self._combine(other, self.AND)


class QCombination(QNode):
    def __init__(self, operation, children):
        self.operation = operation
        self.children = []
        for node in children:
            # If the child is a combination of the same type, we can merge its
            # children directly into this combinations children
            if isinstance(node, QCombination) and node.operation == operation:
                self.children += node.children
            else:
                self.children.append(node)

    def __repr__(self):
        op = " & " if self.operation is self.AND else " | "
        return "(%s)" % op.join([repr(node) for node in self.children])

    def __bool__(self):
        return bool(self.children)

    def accept(self, visitor) -> Union['QCombination', dict]:
        for i in range(len(self.children)):
            if isinstance(self.children[i], QNode):
                self.children[i] = self.children[i].accept(visitor)

        return visitor.prepare_combination(self)

    @property
    def empty(self):
        return not bool(self.children)

    def __eq__(self, other):
        return (
            self.__class__ == other.__class__
            and self.operation == other.operation
            and self.children == other.children
        )


class Q(QNode):
    """A simple query object, used in a query tree to build up more complex
    query structures.
    """

    def __init__(self, **query):
        self.query = query

    def __repr__(self):
        return "Q(**%s)" % repr(self.query)

    def __bool__(self):
        return bool(self.query)

    def __eq__(self, other):
        return self.__class__ == other.__class__ and self.query == other.query

    def accept(self, visit: 'QNodeVisitor') -> Union['Q', dict]:
        return visit.visit_query(self)

    @property
    def empty(self) -> bool:
        return not bool(self.query)


def generate_basic_query(
    manager: 'ODMManager',
    query: dict,
    with_validate_document_fields: bool = True,
) -> dict:
    query_params: dict = {}
    for query_field, value in query.items():
        field, *extra_params = query_field.split("__")
        inners, extra_params = manager._parse_extra_params(extra_params)
        if with_validate_document_fields and not manager._validate_field(field):
            continue
        query_field_name = manager.document.__mapping_query_fields__[field]
        extra = ExtraQueryMapper(manager.document, field).query(extra_params, value)
        if extra:
            value = extra[field]
        elif field == '_id':
            value = validate_object_id(manager.document, value)
        else:
            value = (
                validate_field_value(manager.document, field, value)
                if not inners
                else value
            )
        if inners:
            query_field_name = f'{query_field_name}.{".".join(i for i in inners)}'
        if (
            extra
            and query_field_name in query_params
            and ('__gt' in query_field or '__lt' in query_field)
        ):
            query_params[query_field_name].update(value)
        else:
            query_params[query_field_name] = value
    return query_params



import warnings
from typing import (
    Type,
    Dict,
    Dict,
    Optional,
    Tuple,
    Any,
    TYPE_CHECKING,
    Set,
    cast,
)
from enum import Enum

from pydantic.fields import ModelField
from pydantic.schema import (
    TypeModelOrEnum,
    TypeModelSet,
    field_schema,
    SkipField,
    ROOT_KEY,
    lenient_issubclass,
    default_ref_template,
    enum_process_schema,
    get_model,
    get_flat_models_from_model,
    get_model_name_map,
    get_schema_ref,
)

if TYPE_CHECKING:
    from .document import Document

__all__ = ('model_type_schema', 'model_process_schema', 'model_schema')


def model_type_schema(
    model: Type['Document'],
    *,
    by_alias: bool,
    model_name_map: Dict[TypeModelOrEnum, str],
    ref_template: str,
    ref_prefix: Optional[str] = None,
    known_models: TypeModelSet,
) -> Tuple[Dict[str, Any], Dict[str, Any], Set[str]]:
    """
    You probably should be using ``model_schema()``, this function is indirectly used by that function.
    Take a single ``model`` and generate the schema for its type only, not including additional
    information as title, etc. Also return additional schema definitions, from sub-models.
    """
    properties = {}
    required = []
    definitions: Dict[str, Any] = {}
    nested_models: Set[str] = set()
    all_fields: Dict[str, ModelField] = {
        **model.__fields__,
        **{
            k: cast(ModelField, m.model_field)
            for k, m in model.__motordantic_computed_fields__.items()
        },
    }

    for k, f in all_fields.items():
        try:
            f_schema, f_definitions, f_nested_models = field_schema(
                f,
                by_alias=by_alias,
                model_name_map=model_name_map,
                ref_prefix=ref_prefix,
                ref_template=ref_template,
                known_models=known_models,
            )
        except SkipField as skip:
            warnings.warn(skip.message, UserWarning)
            continue
        definitions.update(f_definitions)
        nested_models.update(f_nested_models)
        if by_alias:
            properties[f.alias] = f_schema
            if f.required:
                required.append(f.alias)
        else:
            properties[k] = f_schema
            if f.required:
                required.append(k)
    if ROOT_KEY in properties:
        out_schema = properties[ROOT_KEY]
        out_schema['title'] = model.__config__.title or model.__name__
    else:
        out_schema = {'type': 'object', 'properties': properties}
        if required:
            out_schema['required'] = required
    if model.__config__.extra == 'forbid':
        out_schema['additionalProperties'] = False
    return out_schema, definitions, nested_models


def model_process_schema(
    model: TypeModelOrEnum,
    *,
    by_alias: bool = True,
    model_name_map: Dict[TypeModelOrEnum, str],
    ref_prefix: Optional[str] = None,
    ref_template: str = default_ref_template,
    known_models: Optional[TypeModelSet] = None,
    field: Optional[ModelField] = None,
) -> Tuple[Dict[str, Any], Dict[str, Any], Set[str]]:
    """
    Used by ``model_schema()``, you probably should be using that function.

    Take a single ``model`` and generate its schema. Also return additional schema definitions, from sub-models. The
    sub-models of the returned schema will be referenced, but their definitions will not be included in the schema. All
    the definitions are returned as the second value.
    """
    from inspect import getdoc, signature

    known_models = known_models or set()
    if lenient_issubclass(model, Enum):
        model = cast(Type[Enum], model)
        s = enum_process_schema(model, field=field)
        return s, {}, set()
    model = cast(Type['Document'], model)
    s = {'title': model.__config__.title or model.__name__}
    doc = getdoc(model)
    if doc:
        s['description'] = doc
    known_models.add(model)
    m_schema, m_definitions, nested_models = model_type_schema(
        model,
        by_alias=by_alias,
        model_name_map=model_name_map,
        ref_prefix=ref_prefix,
        ref_template=ref_template,
        known_models=known_models,
    )
    s.update(m_schema)
    schema_extra = model.__config__.schema_extra
    if callable(schema_extra):
        if len(signature(schema_extra).parameters) == 1:
            schema_extra(s)
        else:
            schema_extra(s, model)
    else:
        s.update(schema_extra)
    return s, m_definitions, nested_models


def model_schema(
    model: Type['Document'],
    by_alias: bool = True,
    ref_prefix: Optional[str] = None,
    ref_template: str = default_ref_template,
) -> Dict[str, Any]:
    """
    Generate a JSON Schema for one model. With all the sub-models defined in the ``definitions`` top-level
    JSON key.

    :param model: a Pydantic model (a class that inherits from BaseModel)
    :param by_alias: generate the schemas using the aliases defined, if any
    :param ref_prefix: the JSON Pointer prefix for schema references with ``$ref``, if None, will be set to the
      default of ``#/definitions/``. Update it if you want the schemas to reference the definitions somewhere
      else, e.g. for OpenAPI use ``#/components/schemas/``. The resulting generated schemas will still be at the
      top-level key ``definitions``, so you can extract them from there. But all the references will have the set
      prefix.
    :param ref_template: Use a ``string.format()`` template for ``$ref`` instead of a prefix. This can be useful for
      references that cannot be represented by ``ref_prefix`` such as a definition stored in another file. For a
      sibling json file in a ``/schemas`` directory use ``"/schemas/${model}.json#"``.
    :return: dict with the JSON Schema for the passed ``model``
    """
    model = get_model(model)  # type: ignore
    flat_models = get_flat_models_from_model(model)
    model_name_map = get_model_name_map(flat_models)
    model_name = model_name_map[model]
    m_schema, m_definitions, nested_models = model_process_schema(
        model,
        by_alias=by_alias,
        model_name_map=model_name_map,
        ref_prefix=ref_prefix,
        ref_template=ref_template,
    )
    if model_name in nested_models:
        # model_name is in Nested models, it has circular references
        m_definitions[model_name] = m_schema
        m_schema = get_schema_ref(model_name, ref_prefix, ref_template, False)
    if m_definitions:
        m_schema.update({'definitions': m_definitions})
    return m_schema



import os
from typing import Optional

from motor.motor_asyncio import AsyncIOMotorClient
from bson.raw_bson import RawBSONDocument

from .singleton import Singleton
from .manager import ODMManager


class MotordanticConnection(object, metaclass=Singleton):
    __slots__ = (
        'address',
        'database_name',
        'max_pool_size',
        'server_selection_timeout_ms',
        'connect_timeout_ms',
        'socket_timeout_ms',
        'ssl_cert_path',
    )

    _connections: dict = {}

    def __init__(
        self,
        address: str,
        database_name: str,
        max_pool_size: int = 250,
        ssl_cert_path: Optional[str] = None,
        server_selection_timeout_ms: int = 60000,
        connect_timeout_ms: int = 30000,
        socket_timeout_ms: int = 60000,
    ):
        self.address = address
        self.database_name = database_name
        self.max_pool_size = max_pool_size
        self.ssl_cert_path = ssl_cert_path
        self.server_selection_timeout_ms = server_selection_timeout_ms
        self.connect_timeout_ms = connect_timeout_ms
        self.socket_timeout_ms = socket_timeout_ms

    def _init_mongo_connection(self, connect: bool = False) -> AsyncIOMotorClient:  # type: ignore
        connection_params: dict = {
            'host': self.address,
            'connect': connect,
            'serverSelectionTimeoutMS': self.server_selection_timeout_ms,
            'maxPoolSize': self.max_pool_size,
            'connectTimeoutMS': self.connect_timeout_ms,
            'socketTimeoutMS': self.socket_timeout_ms,
        }
        if self.ssl_cert_path:
            connection_params['tlsCAFile'] = self.ssl_cert_path
            connection_params['tlsAllowInvalidCertificates'] = bool(self.ssl_cert_path)
        client = AsyncIOMotorClient(
            **connection_params,
            document_class=RawBSONDocument,
        )
        return client

    def _get_motor_client(self) -> AsyncIOMotorClient:  # type: ignore
        pid = os.getpid()
        if pid in self._connections:
            return self._connections[pid]
        else:
            mongo_connection = self._init_mongo_connection()
            self._connections[os.getpid()] = mongo_connection
            return mongo_connection


def connect(
    address: str,
    database_name: str,
    max_pool_size: int = 100,
    ssl_cert_path: Optional[str] = None,
    server_selection_timeout_ms: int = 60000,
    connect_timeout_ms: int = 30000,
    socket_timeout_ms: int = 60000,
) -> MotordanticConnection:
    """init connection to mongodb

    Args:
        address (str): full connection string
        database_name (str): mongo db name
        max_pool_size (int, optional): max connection pool. Defaults to 100.
        ssl_cert_path (Optional[str], optional): path to ssl cert. Defaults to None.
        server_selection_timeout_ms (int, optional): ServerSelectionTimeoutMS. Defaults to 60000.
        connect_timeout_ms (int, optional): ConnectionTimeoutMS. Defaults to 30000.
        socket_timeout_ms (int, optional): SocketTimeoutMS. Defaults to 60000.

    Returns:
        MotordanticConnection: motordantic connection
    """
    os.environ['MOTORDANTIC_DATABASE'] = database_name
    os.environ['MOTORDANTIC_ADDRESS'] = address
    os.environ['MOTORDANTIC_MAX_POOL_SIZE'] = str(max_pool_size)
    os.environ['MOTORDANTIC_SERVER_SELECTION_TIMOUT_MS'] = str(
        server_selection_timeout_ms
    )
    os.environ['MOTORDANTIC_CONNECT_TIMEOUT_MS'] = str(connect_timeout_ms)
    os.environ['MOTORDANTIC_SOCKET_TIMEOUT_MS'] = str(socket_timeout_ms)
    if ssl_cert_path:
        os.environ['MOTORDANTIC_SSL_CERT_PATH'] = ssl_cert_path
    connection = MotordanticConnection(
        address=address,
        database_name=database_name,
        max_pool_size=max_pool_size,
        server_selection_timeout_ms=server_selection_timeout_ms,
        connect_timeout_ms=connect_timeout_ms,
        socket_timeout_ms=socket_timeout_ms,
        ssl_cert_path=ssl_cert_path,
    )
    ODMManager.use(connection)
    return connection



import uuid

import pytest_asyncio
import pytest

from bson import ObjectId

from motordantic.document import Document
from motordantic.exceptions import MotordanticValidationError


class User(Document):
    id: str
    name: str
    email: str


@pytest_asyncio.fixture(scope='session', autouse=True)
async def innert_users(event_loop):
    yield
    await User.Q.drop_collection(force=True)


@pytest.mark.asyncio
async def test_raw_insert_one(connection):
    with pytest.raises(MotordanticValidationError):
        result = await User.Q.raw_query(
            'insert_one', {'id': str(uuid.uuid4()), 'name': {}, 'email': []}
        )
    result = await User.Q.raw_query(
        'insert_one',
        {'id': str(uuid.uuid4()), 'name': 'first', 'email': 'first@mail.ru'},
    )
    assert isinstance(result.inserted_id, ObjectId)


@pytest.mark.asyncio
async def test_raw_find_one(connection):
    result = await User.Q.raw_query('find_one', {'name': 'first'})
    assert result['name'] == 'first'
    assert result['email'] == 'first@mail.ru'


@pytest.mark.asyncio
async def test_raw_update_one(connection):
    with pytest.raises(MotordanticValidationError):
        result = await User.Q.raw_query(
            'update_one', [{'id': uuid.uuid4(), 'name': {}, 'email': []}]
        )
    result = await User.Q.raw_query(
        'update_one', raw_query=({'name': 'first'}, {'$set': {'name': 'updated'}})
    )

    assert result.modified_count == 1

    modifed_result = await User.Q.find_one(email='first@mail.ru')
    assert modifed_result.name == 'updated'
